import numpy as np
from abc import ABC, abstractmethod
from typing import Union, Dict
import matplotlib.pyplot as plt
from femora.components.Material.materialBase import Material
from femora.components.section.section_base import Section


class LayerBase(ABC):
    """
    Abstract base class for layer definitions in fiber sections with enhanced material handling
    """
    
    def __init__(self, material: Union[int, str, Material]):
        self.material = Section.resolve_material(material)
        if self.material is None:
            raise ValueError("Layer requires a valid material")
        self.validate()

    @abstractmethod
    def plot(self, ax: plt.Axes, material_colors: Dict[str, str], 
             show_layer_line: bool = True, show_fibers: bool = True) -> None:
        """Plot the layer on matplotlib axes"""
        pass
    
    @abstractmethod
    def to_tcl(self) -> str:
        """Generate OpenSees TCL command for this layer"""
        pass
    
    @abstractmethod
    def validate(self) -> None:
        """Validate layer parameters"""
        pass
    
    @abstractmethod
    def get_layer_type(self) -> str:
        """Return the layer type name"""
        pass


class StraightLayer(LayerBase):
    """
    Straight layer of fibers between two points with enhanced material handling
    """
    
    def __init__(self, material: Union[int, str, Material], num_fibers: int, area_per_fiber: float,
                 y1: float, z1: float, y2: float, z2: float):
        try:
            self.num_fibers = int(num_fibers)
            self.area_per_fiber = float(area_per_fiber)
            self.y1 = float(y1)
            self.z1 = float(z1)
            self.y2 = float(y2)
            self.z2 = float(z2)
        except (ValueError, TypeError):
            raise ValueError("Invalid numeric values for straight layer parameters")
        
        super().__init__(material)

    def plot(self, ax: plt.Axes, material_colors: Dict[str, str], 
             show_layer_line: bool = True, show_fibers: bool = True) -> None:
        """Plot straight layer"""
        color = material_colors.get(self.material.user_name, 'blue')
        
        if show_layer_line:
            ax.plot([self.y1, self.y2], [self.z1, self.z2], 
                   color=color, linewidth=2, alpha=0.8, linestyle='--')
        
        if show_fibers:
            if self.num_fibers == 1:
                y_pos = (self.y1 + self.y2) / 2
                z_pos = (self.z1 + self.z2) / 2
                ax.plot(y_pos, z_pos, 's', color=color, markersize=6, alpha=0.8)
            else:
                y_positions = np.linspace(self.y1, self.y2, self.num_fibers)
                z_positions = np.linspace(self.z1, self.z2, self.num_fibers)
                
                for y_pos, z_pos in zip(y_positions, z_positions):
                    ax.plot(y_pos, z_pos, 's', color=color, markersize=6, alpha=0.8)
    
    def validate(self) -> None:
        """Validate straight layer parameters"""
        if self.num_fibers <= 0:
            raise ValueError("Number of fibers must be positive")
        if self.area_per_fiber <= 0:
            raise ValueError("Area per fiber must be positive")
        
        # Check that start and end points are different
        if abs(self.y1 - self.y2) < 1e-12 and abs(self.z1 - self.z2) < 1e-12:
            raise ValueError("Start and end points must be different")
    
    def to_tcl(self) -> str:
        """Generate OpenSees TCL command for straight layer"""
        return f"    layer straight {self.material.tag} {self.num_fibers} {self.area_per_fiber} {self.y1} {self.z1} {self.y2} {self.z2}"
    
    def get_layer_type(self) -> str:
        return "Straight"



class CircularLayer(LayerBase):
    """
    Circular layer of fibers along a circular arc with enhanced material handling
    """
    def __init__(self, material: Union[int, str, Material], num_fibers: int, area_per_fiber: float,
                 y_center: float, z_center: float, radius: float,
                 start_ang: float = 0.0, end_ang: float = None):
        try:
            self.num_fibers = int(num_fibers)
            self.area_per_fiber = float(area_per_fiber)
            self.y_center = float(y_center)
            self.z_center = float(z_center)
            self.radius = float(radius)
            if end_ang is None:
                self.end_ang = 360.0 - 360.0/self.num_fibers
            else:
                self.end_ang = float(end_ang)
            self.start_ang = float(start_ang)
        except (ValueError, TypeError):
            raise ValueError("Invalid numeric values for circular layer parameters")
        super().__init__(material)

    def plot(self, ax: plt.Axes, material_colors: Dict[str, str], 
             show_layer_line: bool = True, show_fibers: bool = True) -> None:
        color = material_colors.get(self.material.user_name, 'blue')
        if show_layer_line:
            angles = np.linspace(np.radians(self.start_ang), np.radians(self.end_ang), 100)
            y_arc = self.y_center + self.radius * np.cos(angles)
            z_arc = self.z_center + self.radius * np.sin(angles)
            ax.plot(y_arc, z_arc, color=color, linewidth=2, alpha=0.8, linestyle='--')
        if show_fibers:
            if self.num_fibers == 1:
                angles = [np.radians(self.start_ang)]
            else:
                angles = np.linspace(np.radians(self.start_ang), np.radians(self.end_ang), self.num_fibers)
            for angle in angles:
                y_fiber = self.y_center + self.radius * np.cos(angle)
                z_fiber = self.z_center + self.radius * np.sin(angle)
                ax.plot(y_fiber, z_fiber, 's', color=color, markersize=6, alpha=0.8)

    def validate(self) -> None:
        if self.num_fibers <= 0:
            raise ValueError("Number of fibers must be positive")
        if self.area_per_fiber <= 0:
            raise ValueError("Area per fiber must be positive")
        if self.radius <= 0:
            raise ValueError("Radius must be positive")
        # Normalize angles
        self.start_ang = self.start_ang % 360.0
        self.end_ang = self.end_ang % 360.0
        if abs(self.start_ang - self.end_ang) < 1e-12:
            raise ValueError("Start and end angles cannot be the same")

    def to_tcl(self) -> str:
        # Use default angles if appropriate
        if abs(self.start_ang) < 1e-12 and abs(self.end_ang - (360.0 - 360.0/self.num_fibers)) < 1e-12:
            return f"    layer circ {self.material.tag} {self.num_fibers} {self.area_per_fiber} {self.y_center} {self.z_center} {self.radius}"
        else:
            return f"    layer circ {self.material.tag} {self.num_fibers} {self.area_per_fiber} {self.y_center} {self.z_center} {self.radius} {self.start_ang} {self.end_ang}"

    def get_layer_type(self) -> str:
        return "Circular"


    def get_arc_length(self) -> float:
        """Calculate arc length of the layer"""
        angle_diff = abs(self.end_ang - self.start_ang)


        if angle_diff > 180:  # Handle wrap-around case


            angle_diff = 360 - angle_diff


        return self.radius * np.radians(angle_diff)


    def is_full_circle(self) -> bool:
        """Check if this layer forms a complete circle"""


        angle_span = abs(self.end_ang - self.start_ang)


        return abs(angle_span - 360.0) < 1e-6 or abs(angle_span) < 1e-6


    def __str__(self) -> str:
        arc_length = self.get_arc_length()
        shape_desc = "Full Circle" if self.is_full_circle() else f"Arc ({self.start_ang:.1f}° to {self.end_ang:.1f}°)"
        return f"Circular Layer: {shape_desc}, R={self.radius}, {self.num_fibers} fibers, arc length={arc_length:.3f}, material '{self.material.user_name}'"


from qtpy.QtWidgets import (
    QWidget, QVBoxLayout, QHBoxLayout, QLabel, QLineEdit, 
    QComboBox, QPushButton, QTableWidget, QTableWidgetItem, 
    QDialog, QMessageBox, QHeaderView, QGridLayout,
    QFrame, QTextBrowser, QCheckBox
)
from qtpy.QtCore import Qt
from femora.components.Damping.dampingBase import DampingBase
from femora.components.Region.regionBase import RegionBase, GlobalRegion, ElementRegion, NodeRegion

class DampingSelectorWidget(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        
        self.damping_combo = QComboBox()
        self.refresh_dampings()
        
        self.create_damping_btn = QPushButton("Manage Dampings")
        self.create_damping_btn.clicked.connect(self.open_damping_dialog)
        
        layout.addWidget(QLabel("Damping:"))
        layout.addWidget(self.damping_combo, stretch=1)
        layout.addWidget(self.create_damping_btn)
    
    def refresh_dampings(self):
        self.damping_combo.clear()
        self.damping_combo.addItem("None", None)
        for tag, damping in DampingBase._dampings.items():
            self.damping_combo.addItem(f"{damping.name} (Tag: {tag}, Type: {damping.get_Type()})", damping)
    
    def open_damping_dialog(self):
        from femora.components.Damping.dampingGUI import DampingManagerTab
        dialog = DampingManagerTab(self)
        dialog.exec()
        self.refresh_dampings()
    
    def get_selected_damping(self):
        return self.damping_combo.currentData()
    
    def set_selected_damping(self, damping):
        index = self.damping_combo.findData(damping)
        self.damping_combo.setCurrentIndex(max(index, 0))

class RegionManagerTab(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        layout = QVBoxLayout(self)
        
        # Region type selection
        type_layout = QGridLayout()
        self.region_type_combo = QComboBox()
        self.region_type_combo.addItems(["ElementRegion", "NodeRegion"])
        
        create_region_btn = QPushButton("Create New Region")
        create_region_btn.clicked.connect(self.open_region_creation_dialog)
        
        type_layout.addWidget(QLabel("Region Type:"), 0, 0)
        type_layout.addWidget(self.region_type_combo, 0, 1, 1, 2)
        type_layout.addWidget(create_region_btn, 1, 0, 1, 3)
        
        layout.addLayout(type_layout)
        
        # Regions table
        self.regions_table = QTableWidget()
        self.regions_table.setColumnCount(7)  # Tag, Type, Name, Damping, Edit, Delete
        self.regions_table.setHorizontalHeaderLabels(["Tag", "Type", "Name", "Damping", "Edit", "info" ,"Delete"])
        header = self.regions_table.horizontalHeader()
        header.setSectionResizeMode(QHeaderView.Stretch)
        header.setSectionResizeMode(0, QHeaderView.ResizeToContents)
        # dont show the index column
        self.regions_table.verticalHeader().setVisible(False)
        
        layout.addWidget(self.regions_table)
        
        # Refresh regions button
        refresh_btn = QPushButton("Refresh Regions List")
        refresh_btn.clicked.connect(self.refresh_regions_list)
        layout.addWidget(refresh_btn)
        
        # Initial refresh
        self.refresh_regions_list()

    def refresh_regions_list(self):
        self.regions_table.setRowCount(0)
        regions = RegionBase.get_all_regions()
        
        self.regions_table.setRowCount(len(regions))
        for row, (tag, region) in enumerate(regions.items()):
            # Tag
            tag_item = QTableWidgetItem(str(tag))
            tag_item.setFlags(tag_item.flags() & ~Qt.ItemIsEditable)
            self.regions_table.setItem(row, 0, tag_item)
            
            # Region Type
            type_item = QTableWidgetItem(region.get_type())
            type_item.setFlags(type_item.flags() & ~Qt.ItemIsEditable)
            self.regions_table.setItem(row, 1, type_item)
            
            # Name
            name_item = QTableWidgetItem(region.name)
            name_item.setFlags(name_item.flags() & ~Qt.ItemIsEditable)
            self.regions_table.setItem(row, 2, name_item)
            
            # Damping
            damping_text = region.damping.name if region.damping else "None"
            damping_item = QTableWidgetItem(damping_text)
            damping_item.setFlags(damping_item.flags() & ~Qt.ItemIsEditable)
            self.regions_table.setItem(row, 3, damping_item)
            
            # Edit button
            edit_btn = QPushButton("Edit")
            edit_btn.clicked.connect(lambda checked, r=region: self.open_region_edit_dialog(r))
            self.regions_table.setCellWidget(row, 4, edit_btn)
            
            # Delete button (disabled for global region)
            delete_btn = QPushButton("Delete")
            delete_btn.setEnabled(tag != 0)
            delete_btn.clicked.connect(lambda checked, t=tag: self.delete_region(t))
            self.regions_table.setCellWidget(row, 6, delete_btn)

            # info button
            info_btn = QPushButton("Info")
            info_btn.clicked.connect(lambda checked, r=region: self.show_region_info(r))
            self.regions_table.setCellWidget(row, 5, info_btn)

    def open_region_creation_dialog(self):
        region_type = self.region_type_combo.currentText()
        dialog = RegionCreationDialog(region_type, self)
        if dialog.exec() == QDialog.Accepted:
            self.refresh_regions_list()

    def open_region_edit_dialog(self, region):
        dialog = RegionEditDialog(region, self)
        if dialog.exec() == QDialog.Accepted:
            self.refresh_regions_list()

    def show_region_info(self, region):
        QMessageBox.information(self, f"{region.get_type()} Info", str(region))

    def delete_region(self, tag):
        reply = QMessageBox.question(
            self, 'Delete Region',
            f"Are you sure you want to delete region with tag {tag}?",
            QMessageBox.Yes | QMessageBox.No, QMessageBox.No
        )
        
        if reply == QMessageBox.Yes:
            RegionBase.remove_region(tag)
            self.refresh_regions_list()

class RegionCreationDialog(QDialog):
    def __init__(self, region_type, parent=None):
        super().__init__(parent)
        self.setWindowTitle(f"Create {region_type}")
        self.region_type = region_type
        
        # Main horizontal layout
        main_layout = QHBoxLayout(self)
        
        # Left side - Parameters
        left_widget = QWidget()
        left_layout = QVBoxLayout(left_widget)
        
        # Get region class
        self.region_class = globals()[region_type]
        
        # Parameter inputs
        self.param_inputs = {}
        parameters = self.region_class.get_Parameters()
        
        # Create a grid layout for parameters
        params_group = QGridLayout()
        row = 0
        
        # Add damping selector
        self.damping_selector = DampingSelectorWidget()
        params_group.addWidget(QLabel("Damping:"), row, 0, 1, 3)
        row += 1
        params_group.addWidget(self.damping_selector, row, 0, 1, 3)
        row += 1
        
        # Add other parameters
        for param, input_type in parameters.items():
            if input_type == "checkbox":
                input_field = QCheckBox()
                params_group.addWidget(QLabel(param), row, 0)
                params_group.addWidget(input_field, row, 1, 1, 2)
            else:  # lineEdit
                input_field = QLineEdit()
                params_group.addWidget(QLabel(param), row, 0)
                params_group.addWidget(input_field, row, 1, 1, 2)
            
            self.param_inputs[param] = input_field
            row += 1
        
        left_layout.addLayout(params_group)
        
        # Buttons
        btn_layout = QHBoxLayout()
        create_btn = QPushButton("Create Region")
        create_btn.clicked.connect(self.create_region)
        cancel_btn = QPushButton("Cancel")
        cancel_btn.clicked.connect(self.reject)
        
        btn_layout.addWidget(create_btn)
        btn_layout.addWidget(cancel_btn)
        left_layout.addLayout(btn_layout)
        left_layout.addStretch()
        
        # Add left widget to main layout
        main_layout.addWidget(left_widget)
        
        # Vertical line separator
        line = QFrame()
        line.setFrameShape(QFrame.VLine)
        line.setFrameShadow(QFrame.Sunken)
        main_layout.addWidget(line)
        
        # Right side - Notes
        right_widget = QWidget()
        right_layout = QVBoxLayout(right_widget)
        
        notes_display = QTextBrowser()
        notes_display.setHtml(f"<b>Notes:</b><br>")
        info = self.region_class.getNotes()
        for i, note in enumerate(info["Notes"]):
            notes_display.append(f"<b>{i+1}</b>) {note}")
        notes_display.append(f"<br><b>References:</b>")
        for i, ref in enumerate(info["References"]):
            notes_display.append(f"<b>{i+1}</b>) {ref}")

        right_layout.addWidget(notes_display)
        right_layout.addStretch()
        
        main_layout.addWidget(right_widget)
        
        # Set layout proportions
        main_layout.setStretch(0, 3)
        main_layout.setStretch(2, 2)

    def create_region(self):
        try:
            params = {}
            for param, input_field in self.param_inputs.items():
                if isinstance(input_field, QCheckBox):
                    value = input_field.isChecked()
                else:
                    text = input_field.text().strip()
                    if not text:
                        continue
                    if param in ['elements', 'nodes']:
                        value = [int(x) for x in text.split(',')]
                    elif param in ['element_range', 'node_range']:
                        value = [int(x) for x in text.split(',')][:2]
                    else:
                        value = text
                params[param] = value
            
            damping = self.damping_selector.get_selected_damping()
            region = self.region_class(damping=damping, **params)
            self.accept()

        except ValueError as e:
            QMessageBox.warning(self, "Input Error", str(e))
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e))

class RegionEditDialog(QDialog):
    def __init__(self, region, parent=None):
        super().__init__(parent)
        self.region = region
        self.setWindowTitle(f"Edit {region.get_type()} (Tag: {region.tag})")
        
        # Main horizontal layout
        main_layout = QHBoxLayout(self)
        
        # Left side - Parameters
        left_widget = QWidget()
        left_layout = QVBoxLayout(left_widget)
        
        # Parameter inputs
        self.param_inputs = {}
        parameters = region.__class__.get_Parameters()
        
        # Create a grid layout for parameters
        params_group = QGridLayout()
        row = 0
        
        # Add damping selector
        self.damping_selector = DampingSelectorWidget()
        self.damping_selector.set_selected_damping(region.damping)
        params_group.addWidget(QLabel("Damping:"), row, 0, 1, 3)
        row += 1
        params_group.addWidget(self.damping_selector, row, 0, 1, 3)
        row += 1
        
        # Add other parameters
        for param, input_type in parameters.items():
            if input_type == "checkbox":
                input_field = QCheckBox()
                input_field.setChecked(getattr(region, param, False))
                params_group.addWidget(QLabel(param), row, 0)
                params_group.addWidget(input_field, row, 1, 1, 2)
            else:  # lineEdit
                input_field = QLineEdit()
                value = getattr(region, param, [])
                if value:
                    input_field.setText(','.join(map(str, value)))
                params_group.addWidget(QLabel(param), row, 0)
                params_group.addWidget(input_field, row, 1, 1, 2)
            
            self.param_inputs[param] = input_field
            row += 1
        
        left_layout.addLayout(params_group)
        
        # Buttons
        btn_layout = QHBoxLayout()
        save_btn = QPushButton("Save Changes")
        save_btn.clicked.connect(self.save_changes)
        cancel_btn = QPushButton("Cancel")
        cancel_btn.clicked.connect(self.reject)
        
        btn_layout.addWidget(save_btn)
        btn_layout.addWidget(cancel_btn)
        left_layout.addLayout(btn_layout)
        left_layout.addStretch()
        
        # Add left widget to main layout
        main_layout.addWidget(left_widget)
        
        # Vertical line separator
        line = QFrame()
        line.setFrameShape(QFrame.VLine)
        line.setFrameShadow(QFrame.Sunken)
        main_layout.addWidget(line)
        
        # Right side - Notes
        right_widget = QWidget()
        right_layout = QVBoxLayout(right_widget)
        
        notes_display = QTextBrowser()
        notes_display.setHtml(f"<b>Notes:</b><br>")
        info = region.getNotes()
        for i, note in enumerate(info["Notes"]):
            notes_display.append(f"<b>{i+1}</b>) {note}")
        notes_display.append(f"<br><b>References:</b>")
        for i, ref in enumerate(info["References"]):
            notes_display.append(f"<b>{i+1}</b>) {ref}")
        right_layout.addWidget(notes_display)
        right_layout.addStretch()
        
        main_layout.addWidget(right_widget)
        
        # Set layout proportions
        main_layout.setStretch(0, 3)
        main_layout.setStretch(2, 2)

    def save_changes(self):
        try:
            # Update damping
            self.region.damping = self.damping_selector.get_selected_damping()
            
            # Update other parameters
            params = {}
            for param, input_field in self.param_inputs.items():
                if isinstance(input_field, QCheckBox):
                    setattr(self.region, param, input_field.isChecked())
                else:
                    text = input_field.text().strip()
                    if text:
                        if param in ['elements', 'nodes']:
                            value = [int(x) for x in text.split(',')]
                        elif param in ['element_range', 'node_range']:
                            value = [int(x) for x in text.split(',')][:2]
                        else:
                            value = text
                        setattr(self.region, param, value)
            
            self.accept()

        except ValueError as e:
            QMessageBox.warning(self, "Input Error", str(e))
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e))

if __name__ == "__main__":
    import sys
    from qtpy.QtWidgets import QApplication
    from femora.components.Damping.dampingBase import RayleighDamping, ModalDamping
    
    # Create some sample dampings
    RayleighDamping1 = RayleighDamping(alphaM=0.1, betaK=0.2, betaKInit=0.3, betaKComm=0.4)
    RayleighDamping2 = RayleighDamping(alphaM=0.5, betaK=0.6, betaKInit=0.7, betaKComm=0.8)
    ModalDamping1 = ModalDamping(numberofModes=2, dampingFactors="0.1,0.2")
    
    # Initialize global region
    global_region = GlobalRegion(damping=RayleighDamping1)
    
    # Create some sample regions
    element_region = ElementRegion(damping=ModalDamping1, elements=[1, 2, 3], element_only=True)
    node_region = NodeRegion(damping=RayleighDamping2, nodes=[1, 2, 3, 4])
    
    # Create and show the application
    app = QApplication(sys.argv)
    window = RegionManagerTab()
    window.show()
    sys.exit(app.exec_())


from abc import ABC, abstractmethod
from femora.components.Material.materialBase import MaterialManager

class Action(ABC):
    """
    Abstract base class for all actions in the DRM_GUI.

    This class serves as a blueprint for creating specific actions
    that can be converted into a TCL script representation.
    """

    @abstractmethod
    def to_tcl(self) -> str:
        """
        Convert the action to its TCL script representation.

        This method must be implemented by all subclasses to define
        how the action is represented in TCL script format.

        Returns:
            str: A string containing the TCL script representation of the action.
        """
        raise NotImplementedError("Subclasses must implement the 'to_tcl' method.")



class wipe(Action):
    """
    Action to wipe the current mesh.

    This action clears the current mesh and prepares the system for a new mesh.
    """

    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(wipe, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def to_tcl(self) -> str:
        return "wipe"
    

class wipeAnalysis(Action):
    """
    Action to wipe the current analysis.

    This action clears the current analysis and prepares the system for a new analysis.
    """

    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(wipeAnalysis, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def to_tcl(self) -> str:
        return "wipeAnalysis"
    

class updateMaterialStageToElastic(Action):
    """
    Action to update all materials to elastic stage.
    """
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(updateMaterialStageToElastic, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def to_tcl(self) -> str:
        cmd = ""
        for mat in MaterialManager().get_all_materials().values():
            tmp = mat.updateMaterialStage("Elastic")
            if tmp != "":
                cmd += tmp + "\n"
        return cmd
    
class updateMaterialStageToPlastic(Action):
        """
        Action to update all materials to plastic stage.
        """
        _instance = None

        def __new__(cls, *args, **kwargs):
            if cls._instance is None:
                cls._instance = super(updateMaterialStageToPlastic, cls).__new__(cls, *args, **kwargs)
            return cls._instance

        def to_tcl(self) -> str:
            cmd = ""
            for mat in MaterialManager().get_all_materials().values():
                tmp = mat.updateMaterialStage("Plastic")
                if tmp != "":
                    cmd += tmp + "\n"
            return cmd
        
class reset(Action):
    """
    Action to reset the mesh to its initial state.
    """
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(reset, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def to_tcl(self) -> str:
        return "reset"
    
class loadConst(Action):
    """
    Action to load a constant.
    """
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(loadConst, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def to_tcl(self) -> str:
        return "loadConst"
    

class seTime(Action):
    """
    Action to set the time.
    """

    def __init__(self, pseudo_time: float):
        """
        Initialize the seTime action with a pseudo time.

        Args:
            pseudo_time (float): The pseudo time to set.
        """
        self.pseudo_time = pseudo_time

    def to_tcl(self) -> str:
        return f"setTime {self.pseudo_time}"
    

class exit(Action):
    """
    Action to exit the application.
    """
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(exit, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def to_tcl(self) -> str:
        return "exit"
    

class tcl(Action):
    """
    Action to execute a TCL command.

    This action allows the execution of a TCL command directly.
    """
    def __init__(self, command: str):
        """
        Initialize the tcl action with a command.

        Args:
            command (str): The TCL command to execute.
        """
        self.command = command

    def to_tcl(self) -> str:
        return self.command


class ActionManager:
    """
    A container for all action classes in femora.
    
    This class provides direct access to the action classes rather than instances,
    allowing users to create instances when needed by calling the class.
    """
    _instance = None
    
    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(ActionManager, cls).__new__(cls)
            cls._instance._initialized = False
        return cls._instance
    
    def __init__(self):
        if self._initialized:
            return
        self._initialized = True
        
        # Store the action classes directly
        self.wipe = wipe
        self.wipeAnalysis = wipeAnalysis
        self.updateMaterialStageToElastic = updateMaterialStageToElastic
        self.updateMaterialStageToPlastic = updateMaterialStageToPlastic
        self.reset = reset
        self.loadConst = loadConst
        self.exit = exit
        self.seTime = seTime
        self.tcl = tcl

    
    
    
    @classmethod
    def get_instance(cls):
        """
        Get the singleton instance of ActionManager.
        
        Returns:
            ActionManager: The singleton instance
        """
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance





class PlotterManager:
    """
    A centralized manager for accessing the global plotter.
    This breaks circular import dependencies by providing a clean way to access the plotter.
    """
    _plotter = None

    @classmethod
    def set_plotter(cls, plotter) -> None:
        """
        Set the global plotter instance.
        
        Args:
            plotter: The plotter instance to be set globally
        """
        cls._plotter = plotter

    @classmethod
    def get_plotter(cls):
        """
        Get the global plotter instance.
        
        Returns:
            The global plotter instance
        
        Raises:
            RuntimeError: If no plotter has been set
        """
        if cls._plotter is None:
            raise RuntimeError("Plotter has not been initialized. "
                             "Ensure MainWindow is created before accessing the plotter.")
        return cls._plotter

    @classmethod
    def remove_plotter(cls)->None:
        """
        remove the global plotter instance.
        Useful for testing or resetting the application state.
        """
        cls._plotter = None
    
    @classmethod
    def clear_plotter(cls)->None:
        """
        Clear the plotter.
        """
        cls._plotter.clear()


from typing import List, Dict, Optional, Union, Type
from .base import AnalysisComponent
from abc import ABC, abstractmethod


class ConstraintHandler(AnalysisComponent):
    """
    Base abstract class for constraint handlers, which determine how the constraint equations are enforced
    in the system of equations
    """
    _handlers = {}  # Class-level dictionary to store handler types
    _created_handlers = {}  # Class-level dictionary to track all created handlers
    _next_tag = 1  # Class variable to track the next tag to assign
    
    def __init__(self, handler_type: str):
        """
        Initialize a constraint handler
        
        Args:
            handler_type (str): Type of the constraint handler
        """
        self.tag = ConstraintHandler._next_tag
        ConstraintHandler._next_tag += 1
        self.handler_type = handler_type
        
        # Register this handler in the class-level tracking dictionary
        ConstraintHandler._created_handlers[self.tag] = self
    
    @staticmethod
    def register_handler(name: str, handler_class: Type['ConstraintHandler']):
        """Register a constraint handler type"""
        ConstraintHandler._handlers[name.lower()] = handler_class
    
    @staticmethod
    def create_handler(handler_type: str, **kwargs) -> 'ConstraintHandler':
        """Create a constraint handler of the specified type"""
        handler_type = handler_type.lower()
        if handler_type not in ConstraintHandler._handlers:
            raise ValueError(f"Unknown constraint handler type: {handler_type}")
        return ConstraintHandler._handlers[handler_type](**kwargs)
    
    @staticmethod
    def get_available_types() -> List[str]:
        """Get available constraint handler types"""
        return list(ConstraintHandler._handlers.keys())
    
    @classmethod
    def get_handler(cls, tag: int) -> 'ConstraintHandler':
        """
        Retrieve a specific handler by its tag.
        
        Args:
            tag (int): The tag of the handler
        
        Returns:
            ConstraintHandler: The handler with the specified tag
        
        Raises:
            KeyError: If no handler with the given tag exists
        """
        if tag not in cls._created_handlers:
            raise KeyError(f"No constraint handler found with tag {tag}")
        return cls._created_handlers[tag]


    @classmethod
    def get_all_handlers(cls) -> Dict[int, 'ConstraintHandler']:
        """
        Retrieve all created handlers.
        
        Returns:
            Dict[int, ConstraintHandler]: A dictionary of all handlers, keyed by their unique tags
        """
        return cls._created_handlers
    
    @classmethod
    def clear_all(cls) -> None:
        """
        Clear all handlers and reset tags.
        """
        cls._created_handlers.clear()
        cls._next_tag = 1
    
    @abstractmethod
    def get_values(self) -> Dict[str, Union[str, int, float, bool]]:
        """
        Get the parameters defining this handler
        
        Returns:
            Dict[str, Union[str, int, float, bool]]: Dictionary of parameter values
        """
        pass

    @classmethod
    def _reassign_tags(cls) -> None:
        """
        Reassign tags to all handlers sequentially starting from 1.
        """
        new_handlers = {}
        for idx, handler in enumerate(sorted(cls._created_handlers.values(), key=lambda h: h.tag), start=1):
            handler.tag = idx
            new_handlers[idx] = handler
        cls._created_handlers = new_handlers
        cls._next_tag = len(cls._created_handlers) + 1

    @classmethod
    def remove_handler(cls, tag: int) -> None:
        """
        Delete a handler by its tag and re-tag all remaining handlers sequentially.
        
        Args:
            tag (int): The tag of the handler to delete
        """
        if tag in cls._created_handlers:
            del cls._created_handlers[tag]
            cls._reassign_tags()


class PlainConstraintHandler(ConstraintHandler):
    """
    Plain constraint handler, does not follow the constraint definitions across the model evolution
    """
    def __init__(self):
        super().__init__("Plain")
    
    def to_tcl(self) -> str:
        return "constraints Plain"
    
    def get_values(self) -> Dict[str, Union[str, int, float, bool]]:
        return {}


class TransformationConstraintHandler(ConstraintHandler):
    """
    Transformation constraint handler, performs static condensation of the constraint degrees of freedom
    """
    def __init__(self):
        super().__init__("Transformation")
    
    def to_tcl(self) -> str:
        return "constraints Transformation"
    
    def get_values(self) -> Dict[str, Union[str, int, float, bool]]:
        return {}


class PenaltyConstraintHandler(ConstraintHandler):
    """
    Penalty constraint handler, uses penalty numbers to enforce constraints
    """
    def __init__(self, alpha_s: float, alpha_m: float):
        super().__init__("Penalty")
        self.alpha_s = alpha_s
        self.alpha_m = alpha_m
    
    def to_tcl(self) -> str:
        return f"constraints Penalty {self.alpha_s} {self.alpha_m}"
    
    def get_values(self) -> Dict[str, Union[str, int, float, bool]]:
        return {
            "alpha_s": self.alpha_s,
            "alpha_m": self.alpha_m
        }


class LagrangeConstraintHandler(ConstraintHandler):
    """
    Lagrange multipliers constraint handler, uses Lagrange multipliers to enforce constraints
    """
    def __init__(self, alpha_s: float = 1.0, alpha_m: float = 1.0):
        super().__init__("Lagrange")
        self.alpha_s = alpha_s
        self.alpha_m = alpha_m
    
    def to_tcl(self) -> str:
        return f"constraints Lagrange {self.alpha_s} {self.alpha_m}"
    
    def get_values(self) -> Dict[str, Union[str, int, float, bool]]:
        return {
            "alpha_s": self.alpha_s,
            "alpha_m": self.alpha_m
        }


class AutoConstraintHandler(ConstraintHandler):
    """
    Automatic constraint handler, automatically selects the penalty value for compatibility constraints
    """
    def __init__(self, verbose: bool = False, auto_penalty: Optional[float] = None, 
                 user_penalty: Optional[float] = None):
        super().__init__("Auto")
        self.verbose = verbose
        self.auto_penalty = auto_penalty
        self.user_penalty = user_penalty
    
    def to_tcl(self) -> str:
        cmd = "constraints Auto"
        if self.verbose:
            cmd += " -verbose"
        if self.auto_penalty is not None:
            cmd += f" -autoPenalty {self.auto_penalty}"
        if self.user_penalty is not None:
            cmd += f" -userPenalty {self.user_penalty}"
        return cmd
    
    def get_values(self) -> Dict[str, Union[str, int, float, bool]]:
        return {
            "verbose": self.verbose,
            "auto_penalty": self.auto_penalty,
            "user_penalty": self.user_penalty
        }


class ConstraintHandlerManager:
    """
    Singleton class for managing constraint handlers
    """
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(ConstraintHandlerManager, cls).__new__(cls)
        return cls._instance
    
    def __init__(self):
        self.transfomation = TransformationConstraintHandler
        self.plain = PlainConstraintHandler
        self.penalty = PenaltyConstraintHandler
        self.lagrange = LagrangeConstraintHandler
        self.auto = AutoConstraintHandler
        

    def create_handler(self, handler_type: str, **kwargs) -> ConstraintHandler:
        """Create a new constraint handler"""
        return ConstraintHandler.create_handler(handler_type, **kwargs)

    def get_handler(self, tag: int) -> ConstraintHandler:
        """Get constraint handler by tag"""
        return ConstraintHandler.get_handler(tag)

    def remove_handler(self, tag: int) -> None:
        """Remove constraint handler by tag"""
        ConstraintHandler.remove_handler(tag)

    def get_all_handlers(self) -> Dict[int, ConstraintHandler]:
        """Get all constraint handlers"""
        return ConstraintHandler.get_all_handlers()

    def get_available_types(self) -> List[str]:
        """Get list of available constraint handler types"""
        return ConstraintHandler.get_available_types()
    
    def clear_all(self):
        """Clear all constraint handlers"""  
        ConstraintHandler.clear_all()


# Register all constraint handlers
ConstraintHandler.register_handler('plain', PlainConstraintHandler)
ConstraintHandler.register_handler('transformation', TransformationConstraintHandler)
ConstraintHandler.register_handler('penalty', PenaltyConstraintHandler)
ConstraintHandler.register_handler('lagrange', LagrangeConstraintHandler)
ConstraintHandler.register_handler('auto', AutoConstraintHandler)


"""
This module defines various mesh part instances with the base class of MeshPart.
It includes the StructuredRectangular3D class, which represents a 3D 
structured rectangular mesh part.It provides functionality to 
initialize, generate, and validate a structured rectangular mesh grid using 
the PyVista library.

Classes:
    StructuredRectangular3D: A class representing a 3D structured rectangular mesh part.
    CustomRectangularGrid3D: A class representing a 3D custom rectangular grid mesh part.
    GeometricStructuredRectangular3D: A class representing a 3D geometric structured rectangular mesh part.
    CustomMesh: A class representing a custom mesh part that can load from a file or accept an existing PyVista mesh.

Functions:
    generate_mesh: Generates a structured rectangular mesh.
    get_parameters: Returns a list of parameters for the mesh part type.
    validate_parameters: Validates the input parameters for the mesh part.
    is_elemnt_compatible: Returns True if the element is compatible with the mesh part type.
    update_parameters: Updates the mesh part parameters.
    get_Notes: Returns notes for the mesh part type.


Usage:
    This module is intended to be used as part of the MeshMaker lib for 
    defining and manipulating 3D structured rectangular mesh parts.
"""
from typing import Dict, List, Tuple, Union, Optional
from abc import ABC, abstractmethod
import numpy as np
import pyvista as pv
import os
from femora.components.Mesh.meshPartBase import MeshPart, MeshPartRegistry
from femora.components.Element.elementBase import Element
from femora.components.Region.regionBase import RegionBase



class StructuredRectangular3D(MeshPart):
    """
    Structured Rectangular 3D Mesh Part

    Parameters
    ----------
    user_name : str
        Unique user name for the mesh part.
    element : Element
        Associated element.
    region : RegionBase, optional
        Associated region.
    X Min : float
        Minimum X coordinate.
    X Max : float
        Maximum X coordinate.
    Y Min : float
        Minimum Y coordinate.
    Y Max : float
        Maximum Y coordinate.
    Z Min : float
        Minimum Z coordinate.
    Z Max : float
        Maximum Z coordinate.
    Nx Cells : int
        Number of cells in X direction.
    Ny Cells : int
        Number of cells in Y direction.
    Nz Cells : int
        Number of cells in Z direction.
    """
    _compatible_elements = ["stdBrick", "bbarBrick", "SSPbrick", "PML3D"]
    def __init__(self, user_name: str, element: Element, region: RegionBase=None,**kwargs):
        """
        Initialize a 3D Structured Rectangular Mesh Part.

        Parameters
        ----------
        user_name : str
            Unique user name for the mesh part.
        element : Element
            Associated element.
        region : RegionBase, optional
            Associated region.
        X Min : float
            Minimum X coordinate.
        X Max : float
            Maximum X coordinate.
        Y Min : float
            Minimum Y coordinate.
        Y Max : float
            Maximum Y coordinate.
        Z Min : float
            Minimum Z coordinate.
        Z Max : float
            Maximum Z coordinate.
        Nx Cells : int
            Number of cells in X direction.
        Ny Cells : int
            Number of cells in Y direction.
        Nz Cells : int
            Number of cells in Z direction.
        """
        super().__init__(
            category='volume mesh',
            mesh_type='Rectangular Grid',
            user_name=user_name,
            element=element,
            region=region
        )
        kwargs = self.validate_parameters(**kwargs)
        self.params = kwargs if kwargs else {}
        self.generate_mesh()


    def generate_mesh(self) -> pv.UnstructuredGrid:
        """
        Generate a structured rectangular mesh
        
        Returns:
            pv.UnstructuredGrid: Generated mesh
        """
        
        # Extract parameters
        x_min = self.params.get('X Min', 0)
        x_max = self.params.get('X Max', 1)
        y_min = self.params.get('Y Min', 0)
        y_max = self.params.get('Y Max', 1)
        z_min = self.params.get('Z Min', 0)
        z_max = self.params.get('Z Max', 1)
        nx = self.params.get('Nx Cells', 10)
        ny = self.params.get('Ny Cells', 10)
        nz = self.params.get('Nz Cells', 10)
        X = np.linspace(x_min, x_max, nx + 1)
        Y = np.linspace(y_min, y_max, ny + 1)
        Z = np.linspace(z_min, z_max, nz + 1)
        X, Y, Z = np.meshgrid(X, Y, Z, indexing='ij')
        self.mesh = pv.StructuredGrid(X, Y, Z).cast_to_unstructured_grid()
        return self.mesh


    @classmethod
    def get_parameters(cls) -> List[Tuple[str, str]]:
        """
        Get the list of parameters for this mesh part type.
        
        Returns:
            List[str]: List of parameter names
        """
        return [
            ("X Min", "Minimum X coordinate (float)"),
            ("X Max", "Maximum X coordinate (float)"),
            ("Y Min", "Minimum Y coordinate (float)"),
            ("Y Max", "Maximum Y coordinate (float)"),
            ("Z Min", "Minimum Z coordinate (float)"),
            ("Z Max", "Maximum Z coordinate (float)"),
            ("Nx Cells", "Number of cells in X direction (integer)"),
            ("Ny Cells", "Number of cells in Y direction (integer)"),
            ("Nz Cells", "Number of cells in Z direction (integer)")
        ]
    

    @classmethod
    def validate_parameters(cls, **kwargs) -> Dict[str, Union[int, float, str]]:
        """
        Check if the mesh part input parameters are valid.
        
        Returns:
            Dict[str, Union[int, float, str]]: Dictionary of parmaeters with valid values
        """
        valid_params = {}
        for param_name in ['X Min', 'X Max', 'Y Min', 'Y Max', 'Z Min', 'Z Max']:
            if param_name in kwargs:
                try:
                    valid_params[param_name] = float(kwargs[param_name])
                except ValueError:
                    raise ValueError(f"{param_name} must be a float number")
            else:
                raise ValueError(f"{param_name} parameter is required")
        
        for param_name in ['Nx Cells', 'Ny Cells', 'Nz Cells']:
            if param_name in kwargs:
                try:
                    valid_params[param_name] = int(kwargs[param_name])
                except ValueError:
                    raise ValueError(f"{param_name} must be an integer number")
            else:
                raise ValueError(f"{param_name} parameter is required")
            
        if valid_params['X Min'] >= valid_params['X Max']:
            raise ValueError("X Min must be less than X Max")
        if valid_params['Y Min'] >= valid_params['Y Max']:
            raise ValueError("Y Min must be less than Y Max")
        if valid_params['Z Min'] >= valid_params['Z Max']:
            raise ValueError("Z Min must be less than Z Max")
        
        if valid_params['Nx Cells'] <= 0:
            raise ValueError("Nx Cells must be greater than 0")
        if valid_params['Ny Cells'] <= 0:
            raise ValueError("Ny Cells must be greater than 0")
        if valid_params['Nz Cells'] <= 0:
            raise ValueError("Nz Cells must be greater than 0")
        
        return valid_params
    
    @classmethod
    def is_elemnt_compatible(cls, element:str) -> bool:
        """
        Get the list of compatible element types
        Returns:
            List[str]: List of compatible element types
        """
        return element in cls._compatible_elements
    


    def update_parameters(self, **kwargs) -> None:
        """
        Update mesh part parameters
        
        Args:
            **kwargs: Keyword arguments to update
        """
        validated_params = self.validate_parameters(**kwargs)
        self.params = validated_params


    @staticmethod
    def get_Notes() -> Dict[str, Union[str, list]]:
        """
        Get notes for the mesh part type
        
        Returns:
            Dict[str, Union[str, list]]: Dictionary containing notes about the mesh part
        """
        return {
            "description": "Generates a structured 3D rectangular grid mesh with uniform spacing",
            "usage": [
                "Used for creating regular 3D meshes with equal spacing in each direction",
                "Suitable for simple geometries where uniform mesh density is desired",
                "Efficient for problems requiring regular grid structures"
            ],
            "limitations": [
                "Only creates rectangular/cuboid domains",
                "Cannot handle irregular geometries",
                "Uniform spacing in each direction"
            ],
            "tips": [
                "Ensure the number of cells (Nx, Ny, Nz) is appropriate for your analysis",
                "Consider mesh density requirements for accuracy",
                "Check that the domain bounds (Min/Max) cover your area of interest"
            ]
        }

# Register the 3D Structured Rectangular mesh part type
MeshPartRegistry.register_mesh_part_type('Volume mesh', 'Uniform Rectangular Grid', StructuredRectangular3D)


class CustomRectangularGrid3D(MeshPart):
    """
    Custom Rectangular Grid 3D Mesh Part

    Parameters
    ----------
    user_name : str
        Unique user name for the mesh part.
    element : Element
        Associated element.
    region : RegionBase, optional
        Associated region.
    x_coords : list of float (comma separated string)
        List of X coordinates.
    y_coords : list of float (comma separated string)
        List of Y coordinates.
    z_coords : list of float (comma separated string)
        List of Z coordinates.
    """
    _compatible_elements = ["stdBrick", "bbarBrick", "SSPbrick", "PML3D"]
    def __init__(self, user_name: str, element: Element, region: RegionBase=None,**kwargs):
        """
        Initialize a 3D Custom Rectangular Grid Mesh Part.

        Parameters
        ----------
        user_name : str
            Unique user name for the mesh part.
        element : Element
            Associated element.
        region : RegionBase, optional
            Associated region.
        x_coords : list of float (comma separated string)
            List of X coordinates.
        y_coords : list of float (comma separated string)
            List of Y coordinates.
        z_coords : list of float (comma separated string)
            List of Z coordinates.
        """
        super().__init__(
            category='volume mesh',
            mesh_type='Custom Rectangular Grid',
            user_name=user_name,
            element=element,
            region=region
        )
        self.params = kwargs
        self.generate_mesh()

    def generate_mesh(self) -> pv.UnstructuredGrid:
        """
        Generate a custom rectangular grid mesh
        
        Returns:
            pv.UnstructuredGrid: Generated mesh
        """
        x_coords = self.params.get('x_coords', None).split(',')
        y_coords = self.params.get('y_coords', None).split(',')
        z_coords = self.params.get('z_coords', None).split(',')
        x = np.array(x_coords)
        y = np.array(y_coords)
        z = np.array(z_coords)
        X, Y, Z = np.meshgrid(x, y, z, indexing='ij')
        self.mesh = pv.StructuredGrid(X, Y, Z).cast_to_unstructured_grid()
        del x, y, z, X, Y, Z
        return self.mesh
    
    @classmethod
    def get_parameters(cls) -> List[Tuple[str, str]]:
        """
        Get the list of parameters for this mesh part type.
        
        Returns:
            List[str]: List of parameter names
        """
        return [
            ("x_coords", "List of X coordinates (List[float] , comma separated, required)"),
            ("y_coords", "List of Y coordinates (List[float] , comma separated, required)"),
            ("z_coords", "List of Z coordinates (List[float] , comma separated, required)")
        ]
    
    @classmethod
    def validate_parameters(cls, **kwargs) -> Dict[str, Union[int, float, str, List[float]]]:
        """
        Check if the mesh part input parameters are valid.
        
        Returns:
            Dict[str, Union[int, float, str, List[float]]: Dictionary of parmaeters with valid values
        """
        valid_params = {}
        for param_name in ['x_coords', 'y_coords', 'z_coords']:
            if param_name in kwargs:
                try:
                    valid_params[param_name] = [float(x) for x in kwargs[param_name].split(',')]
                    # check if the values are in ascending order
                    if not all(valid_params[param_name][i] < valid_params[param_name][i+1] for i in range(len(valid_params[param_name])-1)):
                        raise ValueError(f"{param_name} must be in ascending order")
                    valid_params[param_name] = kwargs[param_name]
                except ValueError:
                    raise ValueError(f"{param_name} must be a list of float numbers")
            else:
                raise ValueError(f"{param_name} parameter is required")
        return valid_params

    @classmethod
    def is_elemnt_compatible(cls, element:str) -> bool:
        """
        Get the list of compatible element types
        Returns:
            List[str]: List of compatible element types
        """
        return element in ["stdBrick", "bbarBrick", "SSPbrick", "PML3D"]


    def update_parameters(self, **kwargs) -> None:
        """
        Update mesh part parameters
        
        Args:
            **kwargs: Keyword arguments to update
        """
        validated_params = self.validate_parameters(**kwargs)
        self.params = validated_params


    @staticmethod
    def get_Notes() -> Dict[str, Union[str, list]]:
        """
        Get notes for the mesh part type
        
        Returns:
            Dict[str, Union[str, list]]: Dictionary containing notes about the mesh part
        """
        return {
            "description": "Generates a 3D rectangular grid mesh with custom spacing",
            "usage": [
                "Used for creating 3D meshes with variable spacing in each direction",
                "Suitable for problems requiring non-uniform mesh density",
                "Useful when specific grid point locations are needed"
            ],
            "limitations": [
                "Only creates rectangular/cuboid domains",
                "Cannot handle irregular geometries",
                "Requires manual specification of all grid points"
            ],
            "tips": [
                "Provide coordinates as comma-separated lists of float values",
                "Ensure coordinates are in ascending order",
                "Consider gradual transitions in spacing for better numerical results"
            ]
        }


# Register the 3D Structured Rectangular mesh part type
MeshPartRegistry.register_mesh_part_type('Volume mesh', 'Custom Rectangular Grid', CustomRectangularGrid3D)




class GeometricStructuredRectangular3D(MeshPart):
    """
    Geometric Structured Rectangular 3D Mesh Part

    Parameters
    ----------
    user_name : str
        Unique user name for the mesh part.
    element : Element
        Associated element.
    region : RegionBase, optional
        Associated region.
    x_min : float
        Minimum X coordinate.
    x_max : float
        Maximum X coordinate.
    y_min : float
        Minimum Y coordinate.
    y_max : float
        Maximum Y coordinate.
    z_min : float
        Minimum Z coordinate.
    z_max : float
        Maximum Z coordinate.
    nx : int
        Number of cells in X direction.
    ny : int
        Number of cells in Y direction.
    nz : int
        Number of cells in Z direction.
    x_ratio : float, optional
        Ratio of cell increment in X direction (default 1).
    y_ratio : float, optional
        Ratio of cell increment in Y direction (default 1).
    z_ratio : float, optional
        Ratio of cell increment in Z direction (default 1).
    """
    _compatible_elements = ["stdBrick", "bbarBrick", "SSPbrick", "PML3D"]
    def __init__(self, user_name: str, element: Element, region: RegionBase=None,**kwargs):
        """
        Initialize a 3D Geometric Structured Rectangular Mesh Part.

        Parameters
        ----------
        user_name : str
            Unique user name for the mesh part.
        element : Element
            Associated element.
        region : RegionBase, optional
            Associated region.
        x_min : float
            Minimum X coordinate.
        x_max : float
            Maximum X coordinate.
        y_min : float
            Minimum Y coordinate.
        y_max : float
            Maximum Y coordinate.
        z_min : float
            Minimum Z coordinate.
        z_max : float
            Maximum Z coordinate.
        nx : int
            Number of cells in X direction.
        ny : int
            Number of cells in Y direction.
        nz : int
            Number of cells in Z direction.
        x_ratio : float, optional
            Ratio of cell increment in X direction (default 1).
        y_ratio : float, optional
            Ratio of cell increment in Y direction (default 1).
        z_ratio : float, optional
            Ratio of cell increment in Z direction (default 1).
        """
        super().__init__(
            category='volume mesh',
            mesh_type='Geometric Rectangular Grid',
            user_name=user_name,
            element=element,
            region=region
        )
        self.params = kwargs
        self.generate_mesh()

    @staticmethod
    def custom_linspace(start, end, num_elements, ratio=1):
        """
        Generate a sequence of numbers between start and end with specified spacing ratio.

        Parameters:
        - start (float): The starting value of the sequence.
        - end (float): The ending value of the sequence.
        - num_elements (int): The number of elements in the sequence.
        - ratio (float, optional): The ratio of increment between consecutive intervals. Defaults to 1 (linear spacing).

        Returns:
        - numpy.ndarray: The generated sequence.
        """
        if num_elements <= 0:
            raise ValueError("Number of elements must be greater than 0")

        if num_elements == 1:
            return np.array([start, end])
        
        num_intervals = num_elements
        total = end - start
        
        if ratio == 1:
            return np.linspace(start, end, num_elements+1)
        else:            
            # Determine the base increment
            x = total * (1 - ratio) / (1 - ratio**num_intervals)
            # Generate the increments using the ratio
            increments = x * (ratio ** np.arange(num_intervals))
            # Compute the cumulative sum and add the start value
            elements = start + np.cumsum(np.hstack([0, increments]))
            return elements

    def generate_mesh(self) -> pv.UnstructuredGrid:
        """
        Generate a geometric structured rectangular mesh
        
        Returns:
            pv.UnstructuredGrid: Generated mesh
        """
        x_min = self.params.get('x_min', 0)
        x_max = self.params.get('x_max', 1)
        y_min = self.params.get('y_min', 0)
        y_max = self.params.get('y_max', 1)
        z_min = self.params.get('z_min', 0)
        z_max = self.params.get('z_max', 1)
        nx = self.params.get('nx', 10)
        ny = self.params.get('ny', 10)
        nz = self.params.get('nz', 10)
        x_ratio = self.params.get('x_ratio', 1)
        y_ratio = self.params.get('y_ratio', 1)
        z_ratio = self.params.get('z_ratio', 1)
        x = self.custom_linspace(x_min, x_max, nx, x_ratio)
        y = self.custom_linspace(y_min, y_max, ny, y_ratio)
        z = self.custom_linspace(z_min, z_max, nz, z_ratio)
        X, Y, Z = np.meshgrid(x, y, z, indexing='ij')
        self.mesh = pv.StructuredGrid(X, Y, Z).cast_to_unstructured_grid()

        return self.mesh

    @classmethod
    def get_parameters(cls) -> List[Tuple[str, str]]:
        """
        Get the list of parameters for this mesh part type.
        
        Returns:
            List[str]: List of parameter names
        """
        return [
            ("x_min", "Minimum X coordinate (float)"),
            ("x_max", "Maximum X coordinate (float)"),
            ("y_min", "Minimum Y coordinate (float)"),
            ("y_max", "Maximum Y coordinate (float)"),
            ("z_min", "Minimum Z coordinate (float)"),
            ("z_max", "Maximum Z coordinate (float)"),
            ("nx", "Number of cells in X direction (integer)"),
            ("ny", "Number of cells in Y direction (integer)"),
            ("nz", "Number of cells in Z direction (integer)"),
            ("x_ratio", "Ratio of cell increment in X direction (float)"),
            ("y_ratio", "Ratio of cell increment in Y direction (float)"),
            ("z_ratio", "Ratio of cell increment in Z direction (float)")
        ]
    
    @classmethod
    def validate_parameters(cls, **kwargs) -> Dict[str, Union[int, float, str]]:
        """
        Check if the mesh part input parameters are valid.
        
        Returns:
            Dict[str, Union[int, float, str]]: Dictionary of parmaeters with valid values
        """
        valid_params = {}
        for param_name in ['x_min', 'x_max', 'y_min', 'y_max', 'z_min', 'z_max']:
            if param_name in kwargs:
                try:
                    valid_params[param_name] = float(kwargs[param_name])
                except ValueError:
                    raise ValueError(f"{param_name} must be a float number")
            else:
                raise ValueError(f"{param_name} parameter is required")
        
        for param_name in ['nx', 'ny', 'nz']:
            if param_name in kwargs:
                try:
                    valid_params[param_name] = int(kwargs[param_name])
                except ValueError:
                    raise ValueError(f"{param_name} must be an integer number")
            else:
                raise ValueError(f"{param_name} parameter is required")
            
        for param_name in ['x_ratio', 'y_ratio', 'z_ratio']:
            if param_name in kwargs:
                try:
                    valid_params[param_name] = float(kwargs[param_name])
                    # check if the value is greater than 0
                    if valid_params[param_name] <= 0:
                        raise ValueError(f"{param_name} must be greater than 0")
                except ValueError:
                    raise ValueError(f"{param_name} must be a float number")
            else:
                valid_params[param_name] = 1
        
        if valid_params['x_min'] >= valid_params['x_max']:
            raise ValueError("x_min must be less than x_max")
        if valid_params['y_min'] >= valid_params['y_max']:
            raise ValueError("y_min must be less than y_max")
        if valid_params['z_min'] >= valid_params['z_max']:
            raise ValueError("z_min must be less than z_max")
        
        if valid_params['nx'] <= 0:
            raise ValueError("nx must be greater than 0")
        if valid_params['ny'] <= 0:
            raise ValueError("ny must be greater than 0")
        if valid_params['nz'] <= 0:
            raise ValueError("nz must be greater than 0")
        
        return valid_params
    
    @classmethod
    def is_elemnt_compatible(cls, element:str) -> bool:
        """
        Get the list of compatible element types
        Returns:
            List[str]: List of compatible element types
        """
        return element in ["stdBrick", "bbarBrick", "SSPbrick", "PML3D"]
    
    def update_parameters(self, **kwargs) -> None:
        """
        Update mesh part parameters
        
        Args:
            **kwargs: Keyword arguments to update
        """
        validated_params = self.validate_parameters(**kwargs)
        self.params = validated_params

    
    @staticmethod
    def get_Notes() -> Dict[str, Union[str, list]]:
        """
        Get notes for the mesh part type
        
        Returns:
            Dict[str, Union[str, list]]: Dictionary containing notes about the mesh part
        """
        return {
            "description": "Generates a structured 3D rectangular grid mesh with geometric spacing",
            "usage": [
                "Used for creating 3D meshes with variable spacing in each direction",
                "Suitable for problems requiring non-uniform mesh density",
                "Useful when specific grid point locations are needed"
            ],
            "limitations": [
                "Only creates rectangular/cuboid domains",
                "Cannot handle irregular geometries",
                "Requires manual specification of all grid points"
            ],
            "tips": [
                "Provide coordinates as comma-separated lists of float values",
                "Ensure coordinates are in ascending order",
                "Consider gradual transitions in spacing for better numerical results"
            ]
        }
    
# Register the 3D Geometric Structured Rectangular mesh part type
MeshPartRegistry.register_mesh_part_type('Volume mesh', 'Geometric Rectangular Grid', GeometricStructuredRectangular3D)


class ExternalMesh(MeshPart):
    """
    Custom Mesh Part that can load from a file or accept an existing PyVista mesh.

    Parameters
    ----------
    user_name : str
        Unique user name for the mesh part.
    element : Element
        Associated element.
    region : RegionBase, optional
        Associated region.
    mesh : pv.UnstructuredGrid, optional
        Existing PyVista mesh to use.
    filepath : str, optional
        Path to mesh file to load.
    scale : float, optional
        Scale factor for the mesh.
    rotate_x : float, optional
        Rotation angle around X-axis in degrees.
    rotate_y : float, optional
        Rotation angle around Y-axis in degrees.
    rotate_z : float, optional
        Rotation angle around Z-axis in degrees.
    translate_x : float, optional
        Translation along X-axis.
    translate_y : float, optional
        Translation along Y-axis.
    translate_z : float, optional
        Translation along Z-axis.
    transform_args : dict, optional
        Optional transformation arguments dictionary containing any of the above transformation parameters.
    """
    _compatible_elements = ["stdBrick", "bbarBrick", "SSPbrick", "PML3D"]
    
    def __init__(self, user_name: str, element: Element, region: RegionBase=None, **kwargs):
        """
        Initialize a Custom Mesh Part.

        Parameters
        ----------
        user_name : str
            Unique user name for the mesh part.
        element : Element
            Associated element.
        region : RegionBase, optional
            Associated region.
        mesh : pv.UnstructuredGrid, optional
            Existing PyVista mesh to use.
        filepath : str, optional
            Path to mesh file to load.
        scale : float, optional
            Scale factor for the mesh.
        rotate_x : float, optional
            Rotation angle around X-axis in degrees.
        rotate_y : float, optional
            Rotation angle around Y-axis in degrees.
        rotate_z : float, optional
            Rotation angle around Z-axis in degrees.
        translate_x : float, optional
            Translation along X-axis.
        translate_y : float, optional
            Translation along Y-axis.
        translate_z : float, optional
            Translation along Z-axis.
        transform_args : dict, optional
            Optional transformation arguments dictionary containing any of the above transformation parameters.
        """
        super().__init__(
            category='volume mesh',
            mesh_type='Custom Mesh',
            user_name=user_name,
            element=element,
            region=region
        )
        self.params = self.validate_parameters(**kwargs)
        self.generate_mesh()

    def generate_mesh(self) -> pv.UnstructuredGrid:
        """
        Generate a mesh by loading from file or using the provided mesh and apply transformations
        
        Returns:
            pv.UnstructuredGrid: Generated mesh
        """
        if 'mesh' in self.params:
            # Use the provided mesh
            self.mesh = self.params['mesh']
        elif 'filepath' in self.params:
            # Load mesh from file
            self.mesh = pv.read(self.params['filepath'])
        else:
            raise ValueError("Either 'mesh' or 'filepath' parameter is required")
            
        # Apply scale if specified
        if 'scale' in self.params:
            self.mesh.scale(self.params['scale'], inplace=True)
            
        # Apply rotations if specified (in X, Y, Z order)
        if 'rotate_x' in self.params:
            self.mesh.rotate_x(self.params['rotate_x'], inplace=True)
        if 'rotate_y' in self.params:
            self.mesh.rotate_y(self.params['rotate_y'], inplace=True)
        if 'rotate_z' in self.params:
            self.mesh.rotate_z(self.params['rotate_z'], inplace=True)
            
        # Apply translation if specified
        translate_x = self.params.get('translate_x', 0)
        translate_y = self.params.get('translate_y', 0)
        translate_z = self.params.get('translate_z', 0)
        
        # Only translate if at least one component is non-zero
        if translate_x != 0 or translate_y != 0 or translate_z != 0:
            self.mesh.translate([translate_x, translate_y, translate_z], inplace=True)
            
        # Ensure we have an unstructured grid
        if not isinstance(self.mesh, pv.UnstructuredGrid):
            try:
                self.mesh = self.mesh.cast_to_unstructured_grid()
            except Exception as e:
                raise ValueError(f"Failed to convert mesh to unstructured grid: {str(e)}")
            
        return self.mesh

    @classmethod
    def get_parameters(cls) -> List[Tuple[str, str]]:
        """
        Get the list of parameters for this mesh part type.
        
        Returns:
            List[Tuple[str, str]]: List of parameter names and descriptions
        """
        return [
            ("mesh", "Existing PyVista mesh object (pv.UnstructuredGrid or convertible)"),
            ("filepath", "Path to mesh file to load (str)"),
            ("scale", "Scale factor for the mesh (float)"),
            ("rotate_x", "Rotation angle around X-axis in degrees (float)"),
            ("rotate_y", "Rotation angle around Y-axis in degrees (float)"),
            ("rotate_z", "Rotation angle around Z-axis in degrees (float)"),
            ("translate_x", "Translation along X-axis (float)"),
            ("translate_y", "Translation along Y-axis (float)"),
            ("translate_z", "Translation along Z-axis (float)"),
        ]
    
    @classmethod
    def validate_parameters(cls, **kwargs) -> Dict[str, Union[int, float, str, pv.UnstructuredGrid, Dict]]:
        """
        Check if the mesh part input parameters are valid.
        
        Returns:
            Dict[str, Union[int, float, str, pv.UnstructuredGrid, Dict]]: Dictionary of parameters with valid values
        """
        valid_params = {}
        
        # Check if either mesh or filepath is provided
        if 'mesh' in kwargs:
            try:
                # Verify it's a PyVista mesh
                mesh = kwargs['mesh']
                if not isinstance(mesh, (pv.UnstructuredGrid, pv.PolyData, pv.StructuredGrid)):
                    raise ValueError("'mesh' parameter must be a PyVista mesh")
                valid_params['mesh'] = mesh
            except Exception as e:
                raise ValueError(f"Invalid mesh object: {str(e)}")
        elif 'filepath' in kwargs:
            filepath = kwargs['filepath']
            if not isinstance(filepath, str):
                raise ValueError("'filepath' parameter must be a string")
            if not os.path.exists(filepath):
                raise FileNotFoundError(f"Mesh file not found: {filepath}")
            valid_params['filepath'] = filepath
        else:
            raise ValueError("Either 'mesh' or 'filepath' parameter is required")
        
        # Process transformation parameters
        transform_params = ['scale', 'rotate_x', 'rotate_y', 'rotate_z', 
                           'translate_x', 'translate_y', 'translate_z']
        
        # Handle transform_args dict if provided
        if 'transform_args' in kwargs and isinstance(kwargs['transform_args'], dict):
            # Extract parameters from transform_args
            for param in transform_params:
                if param in kwargs['transform_args']:
                    try:
                        valid_params[param] = float(kwargs['transform_args'][param])
                    except ValueError:
                        raise ValueError(f"Transform parameter '{param}' must be a number")
        
        # Direct parameters override those in transform_args
        for param in transform_params:
            if param in kwargs:
                try:
                    valid_params[param] = float(kwargs[param])
                except ValueError:
                    raise ValueError(f"Parameter '{param}' must be a number")
                    
        # Scale must be positive if provided
        if 'scale' in valid_params and valid_params['scale'] <= 0:
            raise ValueError("Scale factor must be greater than 0")
            
        return valid_params
    
    @classmethod
    def is_elemnt_compatible(cls, element:str) -> bool:
        """
        Check if an element type is compatible with this mesh part
        
        Args:
            element (str): Element type name to check
            
        Returns:
            bool: True if compatible, False otherwise
        """
        return element in cls._compatible_elements
    
    def update_parameters(self, **kwargs) -> None:
        """
        Update mesh part parameters
        
        Args:
            **kwargs: Keyword arguments to update
        """
        # Merge with existing parameters to maintain required params
        merged_params = {**self.params, **kwargs}
        validated_params = self.validate_parameters(**merged_params)
        self.params = validated_params
        self.generate_mesh()  # Regenerate the mesh with new parameters

    @staticmethod
    def get_Notes() -> Dict[str, Union[str, list]]:
        """
        Get notes for the mesh part type
        
        Returns:
            Dict[str, Union[str, list]]: Dictionary containing notes about the mesh part
        """
        return {
            "description": "Handles custom meshes imported from files or existing PyVista meshes",
            "usage": [
                "Used for importing pre-generated meshes from external sources",
                "Suitable for complex geometries created in other software",
                "Useful when working with irregular or complex domain shapes",
                "Can transform meshes with scaling, rotation, and translation operations",
                "Right now it only supports one kind of element type, but it can be extended to support more",
                "The compatibility of the element type is not checked, so it is the user's responsibility to check that the element type is compatible with the mesh part type"
            ],
            "limitations": [
                "Quality and compatibility of imported meshes depends on the source",
                "Some mesh types may require conversion to unstructured grid",
                "Not all mesh formats preserve all required metadata"
            ],
            "tips": [
                "Ensure imported meshes have appropriate element types for analysis",
                "Check mesh quality after import (aspect ratio, skewness, etc.)",
                "Parameters can be provided directly or in the transform_args dict",
                "Transformations are applied in this order: scale → rotate → translate",
                "Common file formats include .vtk, .vtu, .stl, .obj, and .ply"
            ]
        }

# Register the Custom Mesh part type under the General mesh category
MeshPartRegistry.register_mesh_part_type('General mesh', 'External mesh', ExternalMesh)


class StructuredLineMesh(MeshPart):
    """
    Structured Line Mesh Part for beam/column elements.
    Creates a grid of line elements with arbitrary normal direction.

    Parameters
    ----------
    user_name : str
        Unique user name for the mesh part.
    element : Element
        Associated beam element (must have section and transformation).
    region : RegionBase, optional
        Associated region.
    base_point_x : float, optional
        Base point X coordinate. Default: 0.0
    base_point_y : float, optional
        Base point Y coordinate. Default: 0.0
    base_point_z : float, optional
        Base point Z coordinate. Default: 0.0
    base_vector_1_x : float, optional
        First base vector X component. Default: 1.0
    base_vector_1_y : float, optional
        First base vector Y component. Default: 0.0
    base_vector_1_z : float, optional
        First base vector Z component. Default: 0.0
    base_vector_2_x : float, optional
        Second base vector X component. Default: 0.0
    base_vector_2_y : float, optional
        Second base vector Y component. Default: 1.0
    base_vector_2_z : float, optional
        Second base vector Z component. Default: 0.0
    normal_x : float, optional
        Normal direction X component. Default: 0.0
    normal_y : float, optional
        Normal direction Y component. Default: 0.0
    normal_z : float, optional
        Normal direction Z component. Default: 1.0
    grid_size_1 : int, optional
        Number of elements in direction 1. Default: 10
    grid_size_2 : int, optional
        Number of elements in direction 2. Default: 10
    spacing_1 : float, optional
        Spacing in direction 1. Default: 1.0
    spacing_2 : float, optional
        Spacing in direction 2. Default: 1.0
    length : float, optional
        Length of each line element along normal. Default: 1.0
    offset_1 : float, optional
        Optional offset in direction 1. Default: 0.0
    offset_2 : float, optional
        Optional offset in direction 2. Default: 0.0
    number_of_lines : int, optional
        Number of line elements along normal direction per grid point. Default: 1
    merge_points : bool, optional
        Whether to merge duplicate points. Default: True
    """
    _compatible_elements = ["DispBeamColumn", "ForceBeamColumn", "ElasticBeamColumn", "NonlinearBeamColumn"]
    
    def __init__(self, user_name: str, element: Element, region: Optional[RegionBase]=None, **kwargs):
        """
        Initialize a Structured Line Mesh Part.

        Parameters
        ----------
        user_name : str
            Unique user name for the mesh part.
        element : Element
            Associated beam element (must have section and transformation).
        region : RegionBase, optional
            Associated region.
        base_point_x : float, optional
            Base point X coordinate. Default: 0.0
        base_point_y : float, optional
            Base point Y coordinate. Default: 0.0
        base_point_z : float, optional
            Base point Z coordinate. Default: 0.0
        base_vector_1_x : float, optional
            First base vector X component. Default: 1.0
        base_vector_1_y : float, optional
            First base vector Y component. Default: 0.0
        base_vector_1_z : float, optional
            First base vector Z component. Default: 0.0
        base_vector_2_x : float, optional
            Second base vector X component. Default: 0.0
        base_vector_2_y : float, optional
            Second base vector Y component. Default: 1.0
        base_vector_2_z : float, optional
            Second base vector Z component. Default: 0.0
        normal_x : float, optional
            Normal direction X component. Default: 0.0
        normal_y : float, optional
            Normal direction Y component. Default: 0.0
        normal_z : float, optional
            Normal direction Z component. Default: 1.0
        grid_size_1 : int, optional
            Number of elements in direction 1. Default: 10
        grid_size_2 : int, optional
            Number of elements in direction 2. Default: 10
        spacing_1 : float, optional
            Spacing in direction 1. Default: 1.0
        spacing_2 : float, optional
            Spacing in direction 2. Default: 1.0
        length : float, optional
            Length of each line element along normal. Default: 1.0
        offset_1 : float, optional
            Optional offset in direction 1. Default: 0.0
        offset_2 : float, optional
            Optional offset in direction 2. Default: 0.0
        number_of_lines : int, optional
            Number of line elements along normal direction per grid point. Default: 1
        merge_points : bool, optional
            Whether to merge duplicate points. Default: True
        """
        super().__init__(
            category='line mesh',
            mesh_type='Structured Line Grid',
            user_name=user_name,
            element=element,
            region=region
        )
        
        # Validate element compatibility
        if not self.is_element_compatible(element):
            raise ValueError(f"Element type '{element.element_type}' is not compatible with line mesh. "
                           f"Must be a beam element with section and transformation.")
        
        kwargs = self.validate_parameters(**kwargs)
        self.params = kwargs if kwargs else {}
        self.generate_mesh()

    def is_element_compatible(self, element: Element) -> bool:
        """
        Check if element is compatible with line mesh
        
        Args:
            element (Element): Element to check
            
        Returns:
            bool: True if compatible, False otherwise
        """
        # Check element type compatibility using base class method
        if not self.is_elemnt_compatible(element.element_type):
            return False
        
        # Must have section and transformation
        if not element.get_section() or not element.get_transformation():
            return False
        
        return True

    def generate_mesh(self) -> pv.UnstructuredGrid:
        """
        Generate a structured line mesh
        
        Returns:
            pv.UnstructuredGrid: Generated line mesh
        """
        import numpy as np
        
        # Extract parameters
        base_point = np.array([
            self.params.get('base_point_x', 0),
            self.params.get('base_point_y', 0),
            self.params.get('base_point_z', 0)
        ])
        
        base_vector_1 = np.array([
            self.params.get('base_vector_1_x', 1),
            self.params.get('base_vector_1_y', 0),
            self.params.get('base_vector_1_z', 0)
        ])
        
        base_vector_2 = np.array([
            self.params.get('base_vector_2_x', 0),
            self.params.get('base_vector_2_y', 1),
            self.params.get('base_vector_2_z', 0)
        ])
        
        normal = np.array([
            self.params.get('normal_x', 0),
            self.params.get('normal_y', 0),
            self.params.get('normal_z', 1)
        ])
        
        grid_size_1 = self.params.get('grid_size_1', 10)
        grid_size_2 = self.params.get('grid_size_2', 10)
        spacing_1 = self.params.get('spacing_1', 1.0)
        spacing_2 = self.params.get('spacing_2', 1.0)
        length = self.params.get('length', 1.0)
        offset_1 = self.params.get('offset_1', 0.0)
        offset_2 = self.params.get('offset_2', 0.0)
        number_of_lines = self.params.get('number_of_lines', 1)
        merge_points = self.params.get('merge_points', True)
        
        # Normalize vectors
        base_vector_1 = base_vector_1 / np.linalg.norm(base_vector_1)
        base_vector_2 = base_vector_2 / np.linalg.norm(base_vector_2)
        normal = normal / np.linalg.norm(normal)
        
        # Generate grid points
        points = []
        lines = []
        point_id = 0
        
        for i in range(grid_size_1 + 1):
            for j in range(grid_size_2 + 1):
                # Calculate grid point
                grid_point = (base_point + 
                            (i * spacing_1 + offset_1) * base_vector_1 +
                            (j * spacing_2 + offset_2) * base_vector_2)
                
                # Create multiple line elements along the normal direction
                for k in range(number_of_lines):
                    # Calculate the start and end points for this line segment
                    # Each line covers 1/number_of_lines of the total length
                    t_start = k / number_of_lines
                    t_end = (k + 1) / number_of_lines
                    
                    line_start = grid_point + t_start * length * normal
                    line_end = grid_point + t_end * length * normal
                    
                    # Add points for this line
                    points.append(line_start)
                    points.append(line_end)
                    
                    # Add line (connect two points)
                    lines.extend([2, point_id, point_id + 1])
                    point_id += 2
        
        # Create PyVista PolyData
        if points:
            points_array = np.array(points)
            poly_mesh = pv.PolyData(points_array, lines=lines)
            
            # Merge points if requested
            if merge_points:
                poly_mesh = poly_mesh.merge_points(tolerance=1e-4,
                                                 inplace=False,
                                                 progress_bar=False)
            
            # Cast to UnstructuredGrid
            self.mesh = poly_mesh.cast_to_unstructured_grid()
        else:
            # Create empty mesh if no points
            self.mesh = pv.PolyData().cast_to_unstructured_grid()
        
        return self.mesh

    @classmethod
    def get_parameters(cls) -> List[Tuple[str, str]]:
        """
        Get the list of parameters for this mesh part type.
        
        Returns:
            List[Tuple[str, str]]: List of (parameter_name, description) tuples
        """
        return [
            ('base_point_x', 'Base point X coordinate'),
            ('base_point_y', 'Base point Y coordinate'),
            ('base_point_z', 'Base point Z coordinate'),
            ('base_vector_1_x', 'First base vector X component'),
            ('base_vector_1_y', 'First base vector Y component'),
            ('base_vector_1_z', 'First base vector Z component'),
            ('base_vector_2_x', 'Second base vector X component'),
            ('base_vector_2_y', 'Second base vector Y component'),
            ('base_vector_2_z', 'Second base vector Z component'),
            ('normal_x', 'Normal direction X component'),
            ('normal_y', 'Normal direction Y component'),
            ('normal_z', 'Normal direction Z component'),
            ('grid_size_1', 'Number of elements in direction 1'),
            ('grid_size_2', 'Number of elements in direction 2'),
            ('spacing_1', 'Spacing in direction 1'),
            ('spacing_2', 'Spacing in direction 2'),
            ('length', 'Length of each line element along normal'),
            ('offset_1', 'Optional offset in direction 1'),
            ('offset_2', 'Optional offset in direction 2'),
            ('number_of_lines', 'Number of line elements along normal direction per grid point'),
            ('merge_points', 'Whether to merge duplicate points (True/False)')
        ]

    @classmethod
    def validate_parameters(cls, **kwargs) -> Dict[str, Union[int, float, str]]:
        """
        Validate the input parameters for the line mesh part.
        
        Args:
            **kwargs: Input parameters
            
        Returns:
            Dict[str, Union[int, float, str]]: Validated parameters
            
        Raises:
            ValueError: If parameters are invalid
        """
        validated_params = {}
        
        # Validate base point coordinates
        for coord in ['base_point_x', 'base_point_y', 'base_point_z']:
            if coord in kwargs:
                try:
                    validated_params[coord] = float(kwargs[coord])
                except (ValueError, TypeError):
                    raise ValueError(f"{coord} must be a valid number")
            else:
                validated_params[coord] = 0.0
        
        # Validate base vectors
        for vec_num in [1, 2]:
            for coord in ['x', 'y', 'z']:
                param_name = f'base_vector_{vec_num}_{coord}'
                if param_name in kwargs:
                    try:
                        validated_params[param_name] = float(kwargs[param_name])
                    except (ValueError, TypeError):
                        raise ValueError(f"{param_name} must be a valid number")
                else:
                    # Default values: base_vector_1 = [1,0,0], base_vector_2 = [0,1,0]
                    if vec_num == 1:
                        validated_params[param_name] = 1.0 if coord == 'x' else 0.0
                    else:
                        validated_params[param_name] = 1.0 if coord == 'y' else 0.0
        
        # Validate normal vector
        for coord in ['normal_x', 'normal_y', 'normal_z']:
            if coord in kwargs:
                try:
                    validated_params[coord] = float(kwargs[coord])
                except (ValueError, TypeError):
                    raise ValueError(f"{coord} must be a valid number")
            else:
                validated_params[coord] = 1.0 if coord == 'normal_z' else 0.0
        
        # Validate grid sizes
        for size_param in ['grid_size_1', 'grid_size_2']:
            if size_param in kwargs:
                try:
                    size = int(kwargs[size_param])
                    if size < 0:
                        raise ValueError(f"{size_param} must be non-negative")
                    validated_params[size_param] = size
                except (ValueError, TypeError):
                    raise ValueError(f"{size_param} must be a non-negative integer")
            else:
                validated_params[size_param] = 10
        
        # Validate spacings
        for spacing_param in ['spacing_1', 'spacing_2']:
            if spacing_param in kwargs:
                try:
                    spacing = float(kwargs[spacing_param])
                    if spacing <= 0:
                        raise ValueError(f"{spacing_param} must be positive")
                    validated_params[spacing_param] = spacing
                except (ValueError, TypeError):
                    raise ValueError(f"{spacing_param} must be a positive number")
            else:
                validated_params[spacing_param] = 1.0
        
        # Validate length
        if 'length' in kwargs:
            try:
                length = float(kwargs['length'])
                if length <= 0:
                    raise ValueError("length must be positive")
                validated_params['length'] = length
            except (ValueError, TypeError):
                raise ValueError("length must be a positive number")
        else:
            validated_params['length'] = 1.0
        
        # Validate offsets
        for offset_param in ['offset_1', 'offset_2']:
            if offset_param in kwargs:
                try:
                    offset = float(kwargs[offset_param])
                    validated_params[offset_param] = offset
                except (ValueError, TypeError):
                    raise ValueError(f"{offset_param} must be a valid number")
            else:
                validated_params[offset_param] = 0.0
        
        # Validate number of lines
        if 'number_of_lines' in kwargs:
            try:
                num_lines = int(kwargs['number_of_lines'])
                if num_lines < 1:
                    raise ValueError("number_of_lines must be at least 1")
                validated_params['number_of_lines'] = num_lines
            except (ValueError, TypeError):
                raise ValueError("number_of_lines must be a positive integer")
        else:
            validated_params['number_of_lines'] = 1
        
        # Validate merge_points
        if 'merge_points' in kwargs:
            merge_val = kwargs['merge_points']
            if isinstance(merge_val, str):
                merge_val = merge_val.lower()
                if merge_val in ['true', '1', 'yes', 'on']:
                    validated_params['merge_points'] = True
                elif merge_val in ['false', '0', 'no', 'off']:
                    validated_params['merge_points'] = False
                else:
                    raise ValueError("merge_points must be True/False or a valid boolean string")
            elif isinstance(merge_val, bool):
                validated_params['merge_points'] = merge_val
            else:
                raise ValueError("merge_points must be a boolean value")
        else:
            validated_params['merge_points'] = True
        
        return validated_params

    def update_parameters(self, **kwargs) -> None:
        """
        Update mesh part parameters
        
        Args:
            **kwargs: Keyword arguments to update
        """
        validated_params = self.validate_parameters(**kwargs)
        self.params.update(validated_params)
        self.generate_mesh()

    @classmethod
    def get_Notes(cls) -> Dict[str, Union[str, List[str]]]:
        """
        Get notes for the line mesh part type.
        
        Returns:
            Dict[str, Union[str, List[str]]]: Notes with description, usage, limitations, and tips
        """
        return {
            "description": "Structured Line Grid creates a regular grid of line elements (beams/columns) with arbitrary normal direction. Each line element extends from a grid point along the specified normal direction.",
            "usage": [
                "Use for creating column grids in buildings",
                "Use for creating beam grids in floor systems", 
                "Use for creating inclined structural elements",
                "Ensure the associated element is a beam type with section and transformation",
                "Specify base vectors to define the grid orientation",
                "Specify normal vector to define the direction of line elements"
            ],
            "limitations": [
                "Only compatible with beam element types",
                "Requires element to have both section and geometric transformation",
                "Grid is limited to rectangular patterns",
                "All lines have the same length"
            ],
            "tips": [
                "Use base_vector_1 and base_vector_2 to define the grid plane orientation",
                "Use normal vector to define the direction of the line elements",
                "Set grid_size_1 and grid_size_2 to control the density of the grid",
                "Use spacing_1 and spacing_2 to control the distance between lines",
                "Use offset_1 and offset_2 to shift the entire grid",
                "Ensure base vectors are perpendicular for best results"
            ]
        }

# Register the Structured Line mesh part type
MeshPartRegistry.register_mesh_part_type('Line mesh', 'Structured Line Grid', StructuredLineMesh)


class SingleLineMesh(MeshPart):
    """
    Single Line Mesh Part for beam/column elements.
    Creates a single line element between two points.

    Parameters
    ----------
    user_name : str
        Unique user name for the mesh part.
    element : Element
        Associated beam element (must have section and transformation).
    region : RegionBase, optional
        Associated region.
    x0 : float, optional
        Start point X coordinate. Default: 0.0
    y0 : float, optional
        Start point Y coordinate. Default: 0.0
    z0 : float, optional
        Start point Z coordinate. Default: 0.0
    x1 : float, optional
        End point X coordinate. Default: 1.0
    y1 : float, optional
        End point Y coordinate. Default: 0.0
    z1 : float, optional
        End point Z coordinate. Default: 0.0
    number_of_lines : int, optional
        Number of line elements to create along the path. Default: 1
    merge_points : bool, optional
        Whether to merge duplicate points. Default: True
    """
    _compatible_elements = ["DispBeamColumn", "ForceBeamColumn", "ElasticBeamColumn", "NonlinearBeamColumn"]
    
    def __init__(self, user_name: str, element: Element, region: Optional[RegionBase]=None, **kwargs):
        """
        Initialize a Single Line Mesh Part.

        Parameters
        ----------
        user_name : str
            Unique user name for the mesh part.
        element : Element
            Associated beam element (must have section and transformation).
        region : RegionBase, optional
            Associated region.
        x0 : float, optional
            Start point X coordinate. Default: 0.0
        y0 : float, optional
            Start point Y coordinate. Default: 0.0
        z0 : float, optional
            Start point Z coordinate. Default: 0.0
        x1 : float, optional
            End point X coordinate. Default: 1.0
        y1 : float, optional
            End point Y coordinate. Default: 0.0
        z1 : float, optional
            End point Z coordinate. Default: 0.0
        number_of_lines : int, optional
            Number of line elements to create along the path. Default: 1
        merge_points : bool, optional
            Whether to merge duplicate points. Default: True
        """
        super().__init__(
            category='line mesh',
            mesh_type='Single Line',
            user_name=user_name,
            element=element,
            region=region
        )
        
        # Validate element compatibility
        if not self.is_element_compatible(element):
            raise ValueError(f"Element type '{element.element_type}' is not compatible with line mesh. "
                           f"Must be a beam element with section and transformation.")
        
        kwargs = self.validate_parameters(**kwargs)
        self.params = kwargs if kwargs else {}
        self.generate_mesh()

    def is_element_compatible(self, element: Element) -> bool:
        """
        Check if element is compatible with line mesh
        
        Args:
            element (Element): Element to check
            
        Returns:
            bool: True if compatible, False otherwise
        """
        # Check element type compatibility using base class method
        if not self.is_elemnt_compatible(element.element_type):
            return False
        
        # Must have section and transformation
        if not element.get_section() or not element.get_transformation():
            return False
        
        return True

    def generate_mesh(self) -> pv.PolyData:
        """
        Generate a single line mesh
        
        Returns:
            pv.PolyData: Generated line mesh
        """
        import numpy as np
        
        # Extract parameters
        x0 = self.params.get('x0', 0.0)
        y0 = self.params.get('y0', 0.0)
        z0 = self.params.get('z0', 0.0)
        x1 = self.params.get('x1', 1.0)
        y1 = self.params.get('y1', 0.0)
        z1 = self.params.get('z1', 0.0)
        number_of_lines = self.params.get('number_of_lines', 1)
        merge_points = self.params.get('merge_points', True)
        
        # Create start and end points
        start_point = np.array([x0, y0, z0])
        end_point = np.array([x1, y1, z1])
        
        # Calculate the direction vector
        direction = end_point - start_point
        
        # Create points array for multiple lines
        points = []
        lines = []
        point_id = 0
        
        for i in range(number_of_lines):
            # Calculate the start and end points for this line segment
            # Each line covers 1/number_of_lines of the total distance
            t_start = i / number_of_lines
            t_end = (i + 1) / number_of_lines
            
            line_start = start_point + t_start * direction
            line_end = start_point + t_end * direction
            
            # Add points for this line
            points.append(line_start)
            points.append(line_end)
            
            # Add line (connect two points)
            lines.extend([2, point_id, point_id + 1])
            point_id += 2
        
        # Create PyVista PolyData
        if points:
            points_array = np.array(points)
            poly_mesh = pv.PolyData(points_array, lines=lines)
            
            # Merge points if requested
            if merge_points:
                poly_mesh = poly_mesh.merge_points(tolerance=1e-4,
                                                 inplace=False,
                                                 progress_bar=False)
            
            # Cast to UnstructuredGrid
            self.mesh = poly_mesh.cast_to_unstructured_grid()
        else:
            # Create empty mesh if no points
            self.mesh = pv.PolyData().cast_to_unstructured_grid()
        
        return self.mesh

    @classmethod
    def get_parameters(cls) -> List[Tuple[str, str]]:
        """
        Get the list of parameters for this mesh part type.
        
        Returns:
            List[Tuple[str, str]]: List of (parameter_name, description) tuples
        """
        return [
            ('x0', 'Start point X coordinate'),
            ('y0', 'Start point Y coordinate'),
            ('z0', 'Start point Z coordinate'),
            ('x1', 'End point X coordinate'),
            ('y1', 'End point Y coordinate'),
            ('z1', 'End point Z coordinate'),
            ('number_of_lines', 'Number of line elements to create along the path. default is 1'),
            ('merge_points', 'Whether to merge duplicate points (True/False). default is True')
        ]

    @classmethod
    def validate_parameters(cls, **kwargs) -> Dict[str, Union[int, float, str]]:
        """
        Validate the input parameters for the single line mesh part.
        
        Args:
            **kwargs: Input parameters
            
        Returns:
            Dict[str, Union[int, float, str]]: Validated parameters
            
        Raises:
            ValueError: If parameters are invalid
        """
        validated_params = {}
        
        # Validate start point coordinates
        for coord in ['x0', 'y0', 'z0']:
            if coord in kwargs:
                try:
                    validated_params[coord] = float(kwargs[coord])
                except (ValueError, TypeError):
                    raise ValueError(f"{coord} must be a valid number")
            else:
                validated_params[coord] = 0.0
        
        # Validate end point coordinates
        for coord in ['x1', 'y1', 'z1']:
            if coord in kwargs:
                try:
                    validated_params[coord] = float(kwargs[coord])
                except (ValueError, TypeError):
                    raise ValueError(f"{coord} must be a valid number")
            else:
                validated_params[coord] = 1.0 if coord == 'x1' else 0.0
        
        # Validate number of lines
        if 'number_of_lines' in kwargs:
            try:
                num_lines = int(kwargs['number_of_lines'])
                if num_lines < 1:
                    raise ValueError("Number of lines must be at least 1")
                validated_params['number_of_lines'] = num_lines
            except (ValueError, TypeError):
                raise ValueError("number_of_lines must be a positive integer")
        else:
            validated_params['number_of_lines'] = 1
        
        # Check that start and end points are different
        start_point = np.array([validated_params['x0'], validated_params['y0'], validated_params['z0']])
        end_point = np.array([validated_params['x1'], validated_params['y1'], validated_params['z1']])
        
        if np.allclose(start_point, end_point):
            raise ValueError("Start and end points cannot be the same")
        
        # Validate merge_points
        if 'merge_points' in kwargs:
            merge_val = kwargs['merge_points']
            if isinstance(merge_val, str):
                merge_val = merge_val.lower().strip()
                if merge_val in ['true', '1', 'yes', 'on']:
                    validated_params['merge_points'] = True
                elif merge_val in ['false', '0', 'no', 'off']:
                    validated_params['merge_points'] = False
                elif merge_val == '':  # Handle empty string as default True
                    validated_params['merge_points'] = True
                else:
                    raise ValueError("merge_points must be True/False or a valid boolean string")
            elif isinstance(merge_val, bool):
                validated_params['merge_points'] = merge_val
            else:
                raise ValueError("merge_points must be a boolean value")
        else:
            validated_params['merge_points'] = True
        
        return validated_params

    @classmethod
    def is_elemnt_compatible(cls, element: str) -> bool:
        """
        Check if element type is compatible with this mesh part
        
        Args:
            element (str): Element type name
            
        Returns:
            bool: True if compatible, False otherwise
        """
        return element.lower() in [elem.lower() for elem in cls._compatible_elements]

    def update_parameters(self, **kwargs) -> None:
        """
        Update mesh part parameters
        
        Args:
            **kwargs: Keyword arguments to update
        """
        validated_params = self.validate_parameters(**kwargs)
        self.params.update(validated_params)
        self.generate_mesh()

    @classmethod
    def get_Notes(cls) -> Dict[str, Union[str, List[str]]]:
        """
        Get notes for the mesh part type
        
        Returns:
            Dict[str, Union[str, List[str]]]: Dictionary containing notes about the mesh part
        """
        return {
            "description": "Creates a single line element between two specified points. Ideal for individual beam or column elements.",
            "usage": [
                "Use for creating individual beam or column elements",
                "Specify start and end points to define the line direction and length",
                "Compatible with beam elements that have section and transformation properties",
                "Suitable for structural analysis requiring individual line elements"
            ],
            "limitations": [
                "Creates only one line element per mesh part",
                "Requires beam element with section and transformation",
                "Cannot create multiple lines or complex geometries"
            ],
            "tips": [
                "Ensure start and end points are different",
                "Use appropriate beam element type for your analysis",
                "Consider the coordinate system when specifying points",
                "Check that the element has proper section and transformation properties"
            ]
        }

# Register the Single Line mesh part type
MeshPartRegistry.register_mesh_part_type('Line mesh', 'Single Line', SingleLineMesh)



# Manager classes for organizing mesh types
class LineMeshManager:
    """Manager class for line mesh types"""
    single_line = SingleLineMesh
    structured_lines = StructuredLineMesh


class VolumeMeshManager:
    """Manager class for volume mesh types"""
    uniform_rectangular_grid = StructuredRectangular3D
    geometric_rectangular_grid = GeometricStructuredRectangular3D
    external_mesh = ExternalMesh


from typing import List, Dict
from .materialBase import Material, MaterialRegistry
from typing import Union


class ElasticIsotropicMaterial(Material):
    def __init__(self, user_name: str = "Unnamed", **kwargs):
        # validate the parameters
        kwargs = self.validate(**kwargs)
        super().__init__('nDMaterial', 'ElasticIsotropic', user_name)
        self.params = kwargs if kwargs else {}

    
    def to_tcl(self):
        param_order = self.get_parameters()
        params_str = " ".join(str(self.params[param]) for param in param_order if param in self.params)

        return f"{self.material_type} ElasticIsotropic {self.tag} {params_str}; # {self.user_name}"
    
    @staticmethod
    def validate(**params) -> Dict[str, Union[float, int, str, None]]:
        # Extract and validate E
        E = params.get("E")
        if E is None:
            raise ValueError("ElasticIsotropicMaterial requires the 'E' parameter.")
        try:
            E = float(E)
            if E <= 0:
                raise ValueError("Elastic modulus 'E' must be positive.")
        except (ValueError, TypeError):
            raise ValueError("Invalid value for 'E'. It must be a positive number.")

        # Extract and validate nu
        nu = params.get("nu")
        if nu is None:
            raise ValueError("ElasticIsotropicMaterial requires the 'nu' parameter.")
        try:
            nu = float(nu)
            if not (0 <= nu < 0.5):
                raise ValueError("Poisson's ratio 'nu' must be in the range [0, 0.5).")
        except (ValueError, TypeError):
            raise ValueError("Invalid value for 'nu'. It must be a number in range [0, 0.5).")

        # Extract and validate rho
        rho = params.get("rho", 0.0)
        try:
            rho = float(rho)
            if rho < 0:
                raise ValueError("Density 'rho' must be non-negative.")
        except (ValueError, TypeError):
            raise ValueError("Invalid value for 'rho'. It must be a non-negative number.")

        return {"E": E, "nu": nu, "rho": rho}
        
    @classmethod 
    def get_parameters(cls) -> List[str]:
        return ["E", "nu", "rho"]
    
    @classmethod
    def get_description(cls) -> List[str]:
        return ['Young\'s modulus', 
                'Poisson\'s ratio', 
                'Mass density of the material']
    


class ElasticUniaxialMaterial(Material):
    def __init__(self, user_name: str = "Unnamed", **kwargs):
        # validate the parameters
        kwargs = self.validate(**kwargs)
        super().__init__('uniaxialMaterial', 'Elastic', user_name)
        self.params = kwargs if kwargs else {}

    def to_tcl(self):
        param_order = self.get_parameters()
        params_str = " ".join(str(self.params[param]) for param in param_order if param in self.params)
        return f"{self.material_type} Elastic {self.tag} {params_str}; # {self.user_name}"
    
    @staticmethod
    def validate(**params) -> Dict[str, Union[float, int, str, None]]:
        # Extract and validate E
        E = params.get("E")
        if E is None:
            raise ValueError("ElasticUniaxialMaterial requires the 'E' parameter.")
        try:
            E = float(E)
            if E <= 0:
                raise ValueError("Elastic modulus 'E' must be positive.")
        except (ValueError, TypeError):
            raise ValueError("Invalid value for 'E'. It must be a positive number.")

        # Extract and validate eta
        eta = params.get("eta", 0.0)
        try:
            eta = float(eta)
            if eta < 0:
                raise ValueError("Damping ratio 'eta' must be non-negative.")
        except (ValueError, TypeError):
            raise ValueError("Invalid value for 'eta'. It must be a non-negative number.")

        # Extract and validate Eneg
        Eneg = params.get("Eneg", E)
        try:
            Eneg = float(Eneg)
            if Eneg <= 0:
                raise ValueError("Negative elastic modulus 'Eneg' must be positive.")
        except (ValueError, TypeError):
            raise ValueError("Invalid value for 'Eneg'. It must be a positive number.")

        return {"E": E, "eta": eta, "Eneg": Eneg}
        
    @classmethod 
    def get_parameters(cls) -> List[str]:
        return ["E", "eta", "Eneg"]
    
    @classmethod
    def get_description(cls) -> List[str]:
        return ['Tangent', 
                'Damping tangent (optional, default=0.0)',
                'Tangent in compression (optional, default=E)']
    




class J2CyclicBoundingSurfaceMaterial(Material):
    def __init__(self, user_name: str = "Unnamed", **kwargs):
        # validate parameters
        kwargs = self.validate(**kwargs)
        super().__init__('nDMaterial', 'J2CyclicBoundingSurface', user_name)
        self.params = kwargs if kwargs else {}

    def to_tcl(self):
        param_order = self.get_parameters()
        params_str = " ".join(str(self.params[param]) for param in param_order if param in self.params)
        return f"{self.material_type} J2CyclicBoundingSurface {self.tag} {params_str}; # {self.user_name}"
    
    @staticmethod
    def validate(**params) -> Dict[str, Union[float, int, str, None]]:
        required_params = ['G', 'K', 'Su', 'Den', 'h', 'm', 'h0', 'chi']
        validated_params = {}
        
        # Check required parameters
        for param in required_params:
            value = params.get(param)
            if value is None:
                raise ValueError(f"J2CyclicBoundingSurfaceMaterial requires the '{param}' parameter.")
            
            try:
                value = float(value)
            except (ValueError, TypeError):
                raise ValueError(f"Invalid value for '{param}'. It must be a number.")
            
            # Specific validations
            if param in ['G', 'K', 'Su'] and value <= 0:
                raise ValueError(f"'{param}' must be positive.")
            if param == 'Den' and value < 0:
                raise ValueError("Mass density 'Den' must be non-negative.")
            
            validated_params[param] = value
            
        # Optional parameter
        beta = params.get('beta', 0.5)  # Default value
        try:
            beta = float(beta)
            if not (0 <= beta <= 1):
                raise ValueError("Integration variable 'beta' must be in range [0, 1].")
        except (ValueError, TypeError):
            raise ValueError("Invalid value for 'beta'. It must be a number in range [0, 1].")
        
        validated_params['beta'] = beta
        
        return validated_params
    
    @classmethod
    def get_parameters(cls) -> List[str]:
        return ['G', 'K', 'Su', 'Den', 'h', 'm', 'h0', 'chi', 'beta']
    
    @classmethod
    def get_description(cls) -> List[str]:
        return ['Shear modulus', 
                'Bulk modulus',
                'Undrained shear strength',
                'Mass density',
                'Hardening parameter',
                'Hardening exponent',
                'Initial hardening parameter',
                'Initial damping (viscous). chi = 2*dr_o/omega (dr_o = damping ratio at zero strain, omega = angular frequency)',
                'Integration variable (0 = explicit, 1 = implicit, 0.5 = midpoint rule)']
    

    def updateMaterialStage(self, state: str)-> str:
        if state.lower() == 'elastic':
            return "updateMaterialStage -material {self.tag} -stage 0"
        elif state.lower() == 'plastic':
            return "updateMaterialStage -material {self.tag} -stage 1"
        else:
            return ""


class DruckerPragerMaterial(Material):
    def __init__(self, user_name: str = "Unnamed", **kwargs):
        # validate parameters
        kwargs = self.validate(**kwargs)
        super().__init__('nDMaterial', 'DruckerPrager', user_name)
        self.params = kwargs if kwargs else {}

    def to_tcl(self):
        param_order = self.get_parameters()
        params_str = " ".join(str(self.params[param]) for param in param_order if param in self.params)
        return f"{self.material_type} DruckerPrager {self.tag} {params_str}; # {self.user_name}"
    
    @staticmethod
    def validate(**params) -> Dict[str, Union[float, int, str, None]]:
        required_params = ['k', 'G', 'sigmaY', 'rho']
        validated_params = {}
        
        # Check required parameters
        for param in required_params:
            value = params.get(param)
            if value is None:
                raise ValueError(f"DruckerPragerMaterial requires the '{param}' parameter.")
            
            try:
                value = float(value)
            except (ValueError, TypeError):
                raise ValueError(f"Invalid value for '{param}'. It must be a number.")
            
            # Specific validations
            if param in ['k', 'G', 'sigmaY'] and value <= 0:
                raise ValueError(f"'{param}' must be positive.")
            
            validated_params[param] = value
        
        # Optional parameters with specific validations
        optional_params = {
            'rhoBar': {'default': validated_params['rho'], 'min': 0, 'max': validated_params['rho'], 
                      'message': "rhoBar must be in the range [0, rho]"},
            'Kinf': {'default': 0.0, 'min': 0, 'message': "Kinf must be non-negative"},
            'Ko': {'default': 0.0, 'min': 0, 'message': "Ko must be non-negative"},
            'delta1': {'default': 0.0, 'min': 0, 'message': "delta1 must be non-negative"},
            'delta2': {'default': 0.0, 'min': 0, 'message': "delta2 must be non-negative"},
            'H': {'default': 0.0, 'min': 0, 'message': "H must be non-negative"},
            'theta': {'default': 0.0, 'min': 0, 'max': 1, 'message': "theta must be in range [0, 1]"},
            'density': {'default': 0.0, 'min': 0, 'message': "density must be non-negative"},
            'atmPressure': {'default': 101.0, 'min': 0, 'message': "atmPressure must be non-negative"}
        }
        
        for param, constraints in optional_params.items():
            value = params.get(param, constraints['default'])
            try:
                value = float(value)
                if 'min' in constraints and value < constraints['min']:
                    raise ValueError(constraints['message'])
                if 'max' in constraints and value > constraints['max']:
                    raise ValueError(constraints['message'])
            except (ValueError, TypeError):
                raise ValueError(f"Invalid value for '{param}'. It must be a number.")
            
            validated_params[param] = value
        
        return validated_params

    @classmethod 
    def get_parameters(cls) -> List[str]:
        return ['k', 'G', 'sigmaY', 'rho', 'rhoBar', 'Kinf', 'Ko', 'delta1', 'delta2', 'H', 'theta', 'density', 'atmPressure']
    
    @classmethod
    def get_description(cls) -> List[str]:
        return ['Bulk modulus', 
                'Shear modulus',
                'Yield stress',
                'Frictional strength parameter',
                'Controls evolution of plastic volume change: 0 ≤ rhoBar ≤ rho',
                'Nonlinear isotropic strain hardening parameter: Kinf ≥ 0',
                'Nonlinear isotropic strain hardening parameter: Ko ≥ 0',
                'Nonlinear isotropic strain hardening parameter: delta1 ≥ 0',
                'Tension softening parameter: delta2 ≥ 0',
                'Linear strain hardening parameter: H ≥ 0',
                'Controls relative proportions of isotropic and kinematic hardening: 0 ≤ theta ≤ 1',
                'Mass density of the material',
                'Optional atmospheric pressure for update of elastic bulk and shear moduli (default = 101 kPa)']





class PressureDependMultiYieldMaterial(Material):
    """
    OpenSees nD material wrapper for PressureDependMultiYield.

    This material models pressure-sensitive soils with multi-surface (nested)
    plasticity, capturing dilatancy and cyclic mobility. It supports both 2D
    plane strain (nd=2) and 3D (nd=3) analyses.

    Features:
    - Elastic response during gravity/static phase; elastic-plastic in dynamic
      phase when switched via updateMaterialStage.
    - Drucker–Prager type yield surfaces with non-associative flow.
    - Automatic generation of yield surfaces or user-defined γ–Gs backbone.

    See: PressureDependMultiYield material (OpenSees Wiki).
    """
    def __init__(self, user_name: str = "Unnamed", **kwargs):
        """
        Initialize the material with required and optional parameters.

        Required parameters:
        - nd: 2 or 3. Number of spatial dimensions (2=plane strain, 3=3D)
        - rho: Saturated soil mass density
        - refShearModul (Gr): Reference low-strain shear modulus at refPress
        - refBulkModul (Br): Reference bulk modulus at refPress
        - frictionAng (Φ): Friction angle at peak strength, in degrees
        - peakShearStra (γmax): Octahedral shear strain at peak strength
        - refPress (p'r): Reference confining pressure
        - pressDependCoe: Pressure dependence coefficient
        - PTAng (ΦPT): Phase transformation angle, in degrees
        - contrac, dilat1, dilat2: Contraction and dilatancy parameters
        - liquefac1, liquefac2, liquefac3: Liquefaction parameters

        Yield surface definition:
        - noYieldSurf: Number of yield surfaces (< 40). If negative, provide
          'pairs' that define the shear modulus reduction backbone, either as
          a flat list [g1, Gs1, ..., gN, GsN] or a list of tuples
          [(g1, Gs1), ..., (gN, GsN)].

        Optional keyword tail (OpenSees keyword style):
        - e (default 0.6), cs1 (0.9), cs2 (0.02), cs3 (0.7), pa (101.0), c (0.3)

        Raises:
        - ValueError: If any parameter is invalid or inconsistent
        """
        # validate parameters
        kwargs = self.validate(**kwargs)
        super().__init__('nDMaterial', 'PressureDependMultiYield', user_name)
        self.params = kwargs if kwargs else {}

    def to_tcl(self):
        """
        Convert the material to its OpenSees TCL command string.

        Format:
        nDMaterial PressureDependMultiYield tag nd rho Gr Br frictionAng peakShearStra \
          refPress pressDependCoe PTAng contrac dilat1 dilat2 liquefac1 liquefac2 liquefac3 \
          noYieldSurf=<N| -N> [gamma1 Gs1 ... gammaN GsN] \
          e=<val> cs1=<val> cs2=<val> cs3=<val> pa=<val> c=<val>

        Notes:
        - When noYieldSurf is negative, exactly |noYieldSurf| (γ, Gs) pairs must
          be provided and are appended after the noYieldSurf token.

        Returns:
        - str: Complete TCL command with trailing comment of the user_name
        """
        p = self.params
        # Required portion (without user-defined backbone pairs)
        parts = [
            self.material_type,
            'PressureDependMultiYield',
            str(self.tag),
            str(int(p['nd'])),
            str(p['rho']),
            str(p['refShearModul']),
            str(p['refBulkModul']),
            str(p['frictionAng']),
            str(p['peakShearStra']),
            str(p['refPress']),
            str(p['pressDependCoe']),
            str(p['PTAng']),
            str(p['contrac']),
            str(p['dilat1']),
            str(p['dilat2']),
            str(p['liquefac1']),
            str(p['liquefac2']),
            str(p['liquefac3'])
        ]

        # Optional: number of yield surfaces and optional custom backbone pairs
        no_yield = int(p.get('noYieldSurf', 20))
        parts.append(str(no_yield))

        if no_yield < 0:
            pairs = p.get('pairs', [])
            for gamma, gs in pairs:
                parts.append(str(gamma))
                parts.append(str(gs))

        # Optional critical state and atmospheric pressure parameters
        parts.extend([
            str(p['e']),
            str(p['cs1']),
            str(p['cs2']),
            str(p['cs3']),
            str(p['pa']),
            str(p['c'])
        ])

        return " ".join(parts) + f"; # {self.user_name}"

    @staticmethod
    def validate(**params) -> Dict[str, Union[float, int, str, None]]:
        """
        Validate and normalize input parameters.

        - Coerces numeric fields to proper types
        - Checks dimensions, positivity, and ranges (e.g., angles in [0, 90])
        - Validates noYieldSurf and optional 'pairs' structure when negative
        - Applies defaults for e, cs1, cs2, cs3, pa, c

        When noYieldSurf < 0, 'pairs' may be either a flat list [g1, Gs1, ...]
        or a list of tuples [(g1, Gs1), ...].

        Returns:
        - dict: Validated parameters ready for serialization

        Raises:
        - ValueError: On missing params, type issues, range violations, or
          mismatched/invalid backbone pairs
        """
        validated: Dict[str, Union[float, int]] = {}

        # Required parameters
        required = [
            'nd', 'rho', 'refShearModul', 'refBulkModul', 'frictionAng',
            'peakShearStra', 'refPress', 'pressDependCoe', 'PTAng',
            'contrac', 'dilat1', 'dilat2', 'liquefac1', 'liquefac2', 'liquefac3'
        ]

        for key in required:
            value = params.get(key)
            if value is None:
                raise ValueError(f"PressureDependMultiYield requires the '{key}' parameter.")
            try:
                if key == 'nd':
                    value = int(value)
                else:
                    value = float(value)
            except (ValueError, TypeError):
                raise ValueError(f"Invalid value for '{key}'. It must be numeric.")

            # Basic constraints
            if key == 'nd' and value not in (2, 3):
                raise ValueError("'nd' must be 2 (plane strain) or 3 (3D).")
            if key in ['rho', 'refShearModul', 'refBulkModul'] and float(value) <= 0:
                raise ValueError(f"'{key}' must be positive.")
            if key in ['frictionAng', 'PTAng'] and not (0 <= float(value) <= 90):
                raise ValueError(f"'{key}' must be in degrees within [0, 90].")
            if key == 'pressDependCoe' and float(value) < 0:
                raise ValueError("'pressDependCoe' must be non-negative.")
            if key == 'peakShearStra' and float(value) <= 0:
                raise ValueError("'peakShearStra' must be positive.")

            validated[key] = value

        # Optional noYieldSurf and optional custom backbone pairs
        no_yield = params.get('noYieldSurf', 20)
        try:
            no_yield_int = int(no_yield)
        except (ValueError, TypeError):
            raise ValueError("'noYieldSurf' must be an integer.")
        if no_yield_int == 0 or abs(no_yield_int) >= 40:
            raise ValueError("'noYieldSurf' must be non-zero and less than 40 in magnitude.")
        validated['noYieldSurf'] = no_yield_int

        if no_yield_int < 0:
            expected_pairs = abs(no_yield_int)
            pairs_param = params.get('pairs', None)
            if pairs_param is None:
                raise ValueError("When 'noYieldSurf' is negative, provide 'pairs' as a list of (gamma, Gs) or a flat list of length 2N.")

            pairs_list = []
            if isinstance(pairs_param, list):
                # flat list [g1, Gs1, g2, Gs2, ...]
                if len(pairs_param) == 2 * expected_pairs and all(isinstance(x, (int, float)) for x in pairs_param):
                    it = iter(pairs_param)
                    pairs_list = [(float(g), float(gs)) for g, gs in zip(it, it)]
                else:
                    # list of tuples
                    try:
                        pairs_list = [(float(g), float(gs)) for g, gs in pairs_param]
                    except Exception:
                        raise ValueError("'pairs' must be a list of (gamma, Gs) or a flat list of numeric values of length 2N.")
            else:
                raise ValueError("'pairs' must be provided as a list.")

            if len(pairs_list) != expected_pairs:
                raise ValueError(f"Expected {expected_pairs} (gamma, Gs) pairs, got {len(pairs_list)}.")

            for g, gs in pairs_list:
                if g <= 0:
                    raise ValueError("Each gamma must be positive.")
                if not (0 < gs <= 1.0):
                    raise ValueError("Each Gs must be in (0, 1].")

            validated['pairs'] = pairs_list

        # Optional parameters with defaults from OpenSees wiki
        optionals = {
            'e': 0.6,
            'cs1': 0.9,
            'cs2': 0.02,
            'cs3': 0.7,
            'pa': 101.0,
            'c': 0.3,
        }
        for key, default in optionals.items():
            value = params.get(key, default)
            try:
                value = float(value)
            except (ValueError, TypeError):
                raise ValueError(f"Invalid value for '{key}'. It must be numeric.")

            # Basic constraints
            if key in ['e', 'cs1', 'cs2', 'cs3'] and value < 0:
                raise ValueError(f"'{key}' must be non-negative.")
            if key == 'pa' and value <= 0:
                raise ValueError("'pa' must be positive.")
            if key == 'c' and value < 0:
                raise ValueError("'c' must be non-negative.")

            validated[key] = value

        return validated

    @classmethod
    def get_parameters(cls) -> List[str]:
        """
        Parameter keys for GUI/registry.

        Note: 'pairs' applies only when 'noYieldSurf' is negative and allows
        user-defined backbone input in either flat or tuple-list form.
        """
        return [
            'nd', 'rho', 'refShearModul', 'refBulkModul', 'frictionAng',
            'peakShearStra', 'refPress', 'pressDependCoe', 'PTAng',
            'contrac', 'dilat1', 'dilat2', 'liquefac1', 'liquefac2', 'liquefac3', 'noYieldSurf', 'pairs',
            'e', 'cs1', 'cs2', 'cs3', 'pa', 'c'
        ]

    @classmethod
    def get_description(cls) -> List[str]:
        """
        Human-readable descriptions for parameters returned by get_parameters().
        """
        return [
            'Number of dimensions (2 for plane strain, 3 for 3D)',
            'Saturated soil mass density',
            'Reference low-strain shear modulus Gr at refPress',
            'Reference bulk modulus Br at refPress',
            'Friction angle at peak shear strength Φ (degrees)',
            'Octahedral shear strain γmax at peak strength at refPress',
            'Reference mean effective confining pressure p\'r (kPa)',
            'Pressure dependence coefficient',
            'Phase transformation angle ΦPT (degrees)',
            'Contraction parameter',
            'First dilatancy parameter',
            'Second dilatancy parameter',
            'Liquefaction parameter 1',
            'Liquefaction parameter 2',
            'Liquefaction parameter 3',
            'Number of yield surfaces (<40). Use negative with programmatic "pairs" to define γ-Gs backbone',
            'Pairs of γ-Gs backbone',
            'Void ratio e (default 0.6)',
            'Critical state parameter cs1 (default 0.9)',
            'Critical state parameter cs2 (default 0.02)',
            'Critical state parameter cs3 (default 0.7)',
            'Atmospheric pressure pa (kPa, default 101)',
            'Cohesion intercept c (default 0.3)'
        ]

    def updateMaterialStage(self, state: str) -> str:
        """
        Build an OpenSees updateMaterialStage command for this material.

        Parameters:
        - state: 'elastic' -> stage 0; 'plastic' -> stage 1. Any other value
          returns an empty string.

        Returns:
        - str: Command or empty string if state is unrecognized
        """
        if state.lower() == 'elastic':
            return f"updateMaterialStage -material {self.tag} -stage 0"
        elif state.lower() == 'plastic':
            return f"updateMaterialStage -material {self.tag} -stage 1"
        else:
            return ""


# Register material types
MaterialRegistry.register_material_type('nDMaterial', 'ElasticIsotropic', ElasticIsotropicMaterial)
MaterialRegistry.register_material_type('uniaxialMaterial', 'Elastic', ElasticUniaxialMaterial)
MaterialRegistry.register_material_type('nDMaterial', 'J2CyclicBoundingSurface', J2CyclicBoundingSurfaceMaterial)
MaterialRegistry.register_material_type('nDMaterial', 'DruckerPrager', DruckerPragerMaterial)
MaterialRegistry.register_material_type('nDMaterial', 'PressureDependMultiYield', PressureDependMultiYieldMaterial)


class Steel01Material(Material):
    """
    OpenSees uniaxial bilinear steel model (Steel01) with kinematic hardening
    and optional isotropic hardening parameters.

    Reference: OpenSees Wiki – Steel01 Material
    https://opensees.berkeley.edu/wiki/index.php/Steel01_Material
    
    parameters:
        Fy: Yield strength
        E0: Initial elastic tangent
        b: Strain-hardening ratio (post-yield tangent / E0)
        a1, a2: Compression envelope growth parameters
        a3, a4: Tension envelope growth parameters
        If any of a1..a4 is provided, all four must be provided. If none are provided, the TCL output will not include these optional arguments.
    """

    def __init__(self, user_name: str = "Unnamed", **kwargs):
        """
        Initialize Steel01.

        Required parameters:
        - Fy: Yield strength
        - E0: Initial elastic tangent
        - b: Strain-hardening ratio (post-yield tangent / E0)

        Optional isotropic hardening parameters (all-or-none):
        - a1, a2: Compression envelope growth parameters
        - a3, a4: Tension envelope growth parameters
        If any of a1..a4 is provided, all four must be provided. If none are
        provided, the TCL output will not include these optional arguments.

        Raises:
        - ValueError for missing/invalid inputs
        """
        kwargs = self.validate(**kwargs)
        super().__init__('uniaxialMaterial', 'Steel01', user_name)
        self.params = kwargs if kwargs else {}

    def to_tcl(self) -> str:
        """
        Render the OpenSees TCL command for this Steel01 material.

        Format:
        - Without isotropic hardening:
          uniaxialMaterial Steel01 tag Fy E0 b
        - With isotropic hardening (all four provided):
          uniaxialMaterial Steel01 tag Fy E0 b a1 a2 a3 a4
        """
        p = self.params
        ordered = [
            self.material_type,
            'Steel01',
            str(self.tag),
            str(p['Fy']),
            str(p['E0']),
            str(p['b']),
        ]
        # Append isotropic hardening params only if user provided all four
        if all(k in p for k in ['a1', 'a2', 'a3', 'a4']):
            ordered.extend([
                str(p['a1']),
                str(p['a2']),
                str(p['a3']),
                str(p['a4']),
            ])
        return " ".join(ordered) + f"; # {self.user_name}"

    @staticmethod
    def validate(**params) -> Dict[str, Union[float, int, str, None]]:
        """
        Validate parameters for Steel01.
        - Fy, E0 must be positive
        - b must be >= 0
        - a1..a4: either all provided or none; when provided, each must be >= 0
        """
        validated: Dict[str, Union[float, int]] = {}

        # Required
        for key in ['Fy', 'E0', 'b']:
            value = params.get(key)
            if value is None:
                raise ValueError(f"Steel01 requires '{key}'.")
            try:
                value = float(value)
            except (ValueError, TypeError):
                raise ValueError(f"Invalid value for '{key}'. Must be numeric.")
            if key in ['Fy', 'E0'] and value <= 0:
                raise ValueError(f"'{key}' must be positive.")
            if key == 'b' and value < 0:
                raise ValueError("'b' must be non-negative.")
            validated[key] = value

        # Optional isotropic hardening params a1..a4: require all or none
        provided_keys = [k for k in ['a1', 'a2', 'a3', 'a4'] if k in params]
        if len(provided_keys) not in (0, 4):
            raise ValueError("If specifying isotropic parameters, provide all of a1, a2, a3, a4.")
        if len(provided_keys) == 4:
            for key in ['a1', 'a2', 'a3', 'a4']:
                value = params.get(key)
                try:
                    value = float(value)
                except (ValueError, TypeError):
                    raise ValueError(f"Invalid value for '{key}'. Must be numeric.")
                if value < 0:
                    raise ValueError(f"'{key}' must be non-negative.")
                validated[key] = value

        return validated

    @classmethod
    def get_parameters(cls) -> List[str]:
        return ['Fy', 'E0', 'b', 'a1', 'a2', 'a3', 'a4']

    @classmethod
    def get_description(cls) -> List[str]:
        return [
            'Yield strength Fy',
            'Initial elastic tangent E0',
            'Strain-hardening ratio b (post-yield tangent / E0)',
            'Isotropic hardening: compression envelope growth a1',
            'Isotropic hardening parameter a2',
            'Isotropic hardening: tension envelope growth a3',
            'Isotropic hardening parameter a4',
        ]


# Register Steel01
MaterialRegistry.register_material_type('uniaxialMaterial', 'Steel01', Steel01Material)


from PySide6.QtWidgets import (QWidget, QVBoxLayout, QHBoxLayout, QGroupBox, 
                           QPushButton, QTabWidget, QComboBox)
from .rectangularGrid import RectangularGridTab
from .sphericalGrid import SphericalGridTab
from .customGrid import CustomGridTab

class soilSections(QWidget):
    GRID_TYPES = {
        'Rectangular Grid': RectangularGridTab,
        'Spherical Grid': SphericalGridTab,
        'Custom Grid': CustomGridTab
    }

    def __init__(self, parent=None):
        super().__init__(parent)
        self.block_counter = 0
        
        # Main layout
        self.main_layout = QVBoxLayout(self)
        
        # Group box
        self.group_box = QGroupBox('Soil Sections')
        self.group_box_layout = QVBoxLayout(self.group_box)
        
        # Header with grid type selector and Add button
        header_layout = QHBoxLayout()
        
        self.grid_type_selector = QComboBox()
        self.grid_type_selector.addItems(self.GRID_TYPES.keys())
        header_layout.addWidget(self.grid_type_selector)
        
        self.soil_block_button = QPushButton('Add Section')
        self.soil_block_button.clicked.connect(self.add_soil_block)
        header_layout.addWidget(self.soil_block_button)
        
        self.group_box_layout.addLayout(header_layout)
        
        # Tab widget
        self.tab_widget = QTabWidget()
        self.tab_widget.setTabsClosable(True)
        self.tab_widget.tabCloseRequested.connect(self.remove_soil_block)
        self.group_box_layout.addWidget(self.tab_widget)
        
        # Add group box to main layout
        self.main_layout.addWidget(self.group_box)
        
        # Add initial soil block
        self.add_soil_block()

    def add_soil_block(self):
        self.block_counter += 1
        grid_type = self.grid_type_selector.currentText()
        grid_class = self.GRID_TYPES[grid_type]
        
        new_tab = grid_class(block_number=self.block_counter)
        self.tab_widget.addTab(new_tab, f"Block {self.block_counter}")
        self.tab_widget.setCurrentWidget(new_tab)
        
        self.update_tab_names()

    def remove_soil_block(self, index):
        if self.tab_widget.count() > 1:
            tab = self.tab_widget.widget(index)
            if hasattr(tab, 'delete_grid'):
                tab.delete_grid()
            self.tab_widget.removeTab(index)
            self.update_tab_names()

    def update_tab_names(self):
        for i in range(self.tab_widget.count()):
            self.tab_widget.setTabText(i, f"Block {i + 1}")

    def get_soil_blocks_data(self):
        return [self.tab_widget.widget(i).get_data() 
                for i in range(self.tab_widget.count())]


"""
Aggregator Section GUI Dialogs
Uses AggregatorSection class methods instead of hardcoding
"""

from qtpy.QtWidgets import (
    QDialog, QVBoxLayout, QHBoxLayout, QLabel, QLineEdit, QComboBox, QPushButton, QFormLayout, QMessageBox, QGridLayout, QFrame, QTextBrowser, QGroupBox, QWidget
)

from femora.components.section.section_opensees import AggregatorSection
from femora.components.section.section_base import Section
from femora.components.section.section_gui_utils import setup_material_dropdown, get_material_by_combo_selection, set_combo_to_material, setup_uniaxial_material_dropdown


class AggregatorSectionCreationDialog(QDialog):
    """
    Dialog for creating new AggregatorSection sections using class methods
    """
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Create Aggregator Section")
        self.setMinimumSize(500, 400)
        self.response_codes = ['P', 'Mz', 'My', 'Vy', 'Vz', 'T']
        
        self.parameters = AggregatorSection.get_parameters()
        self.descriptions = AggregatorSection.get_description()
        
        main_layout = QHBoxLayout(self)
        
        # Left side - Parameters
        left_widget = QWidget()
        left_layout = QVBoxLayout(left_widget)
        
        # Section name group
        name_group = QGroupBox("Section Identification")
        name_layout = QFormLayout(name_group)
        self.user_name_input = QLineEdit()
        self.user_name_input.setPlaceholderText("Enter unique section name")
        name_layout.addRow("Section Name:", self.user_name_input)
        left_layout.addWidget(name_group)
        
        # Materials group
        materials_group = QGroupBox("Materials by Response Code")
        materials_layout = QGridLayout(materials_group)
        self.material_combos = {}
        for row, code in enumerate(self.response_codes):
            code_label = QLabel(f"{code}:")
            combo = QComboBox()
            setup_uniaxial_material_dropdown(combo, "Select Uniaxial Material")
            materials_layout.addWidget(code_label, row, 0)
            materials_layout.addWidget(combo, row, 1)
            self.material_combos[code] = combo
        left_layout.addWidget(materials_group)
        
        # Base section group (optional)
        base_section_group = QGroupBox("Base Section (Optional)")
        base_section_layout = QFormLayout(base_section_group)
        self.base_section_combo = QComboBox()
        self.base_section_combo.addItem("None", None)
        for tag, section in Section.get_all_sections().items():
            self.base_section_combo.addItem(f"{section.user_name} (Tag: {tag})", section)
        base_section_layout.addRow("Base Section:", self.base_section_combo)
        left_layout.addWidget(base_section_group)
        
        # Buttons
        btn_layout = QHBoxLayout()
        create_btn = QPushButton("Create Section")
        create_btn.clicked.connect(self.create_section)
        cancel_btn = QPushButton("Cancel")
        cancel_btn.clicked.connect(self.reject)
        btn_layout.addWidget(create_btn)
        btn_layout.addWidget(cancel_btn)
        left_layout.addLayout(btn_layout)
        left_layout.addStretch()
        main_layout.addWidget(left_widget)
        
        # Vertical line separator
        line = QFrame()
        line.setFrameShape(QFrame.VLine)
        line.setFrameShadow(QFrame.Sunken)
        main_layout.addWidget(line)
        
        # Right side - Help text from class
        right_widget = QWidget()
        right_layout = QVBoxLayout(right_widget)
        help_display = QTextBrowser()
        help_display.setHtml(AggregatorSection.get_help_text())
        right_layout.addWidget(help_display)
        main_layout.addWidget(right_widget)
        main_layout.setStretch(0, 3)
        main_layout.setStretch(2, 2)

    def create_section(self):
        try:
            user_name = self.user_name_input.text().strip()
            if not user_name:
                QMessageBox.warning(self, "Input Error", "Please enter a section name.")
                return
            try:
                existing_section = AggregatorSection.get_section_by_name(user_name)
                QMessageBox.warning(self, "Input Error", f"Section with name '{user_name}' already exists.")
                return
            except KeyError:
                pass
            # Collect materials
            materials = {}
            for code, combo in self.material_combos.items():
                material = get_material_by_combo_selection(combo)
                if material:
                    materials[code] = material
            if not materials:
                QMessageBox.warning(self, "Input Error", "Please select at least one material for a response code.")
                return
            # Base section
            base_section = self.base_section_combo.currentData()
            params = {"materials": materials}
            if base_section:
                params["base_section"] = base_section
            try:
                validated_params = AggregatorSection.validate_section_parameters(**params)
            except ValueError as e:
                QMessageBox.warning(self, "Validation Error", str(e))
                return
            self.created_section = AggregatorSection(user_name=user_name, **params)
            QMessageBox.information(self, "Success", f"Aggregator Section '{user_name}' created successfully!")
            self.accept()
        except Exception as e:
            QMessageBox.critical(self, "Error", f"Failed to create section: {str(e)}")

class AggregatorSectionEditDialog(QDialog):
    """
    Dialog for editing existing AggregatorSection sections using class methods
    """
    def __init__(self, section, parent=None):
        super().__init__(parent)
        self.response_codes = ['P', 'Mz', 'My', 'Vy', 'Vz', 'T']
        self.section = section
        self.setWindowTitle(f"Edit Aggregator Section: {section.user_name}")
        self.setMinimumSize(800, 600)
        self.parameters = section.get_parameters()
        self.descriptions = section.get_description()
        main_layout = QHBoxLayout(self)
        left_widget = QWidget()
        left_layout = QVBoxLayout(left_widget)
        info_group = QGroupBox("Section Information")
        info_layout = QFormLayout(info_group)
        self.tag_label = QLabel(str(section.tag))
        self.user_name_label = QLabel(section.user_name)
        self.type_label = QLabel(section.section_name)
        info_layout.addRow("Tag:", self.tag_label)
        info_layout.addRow("Name:", self.user_name_label)
        info_layout.addRow("Type:", self.type_label)
        left_layout.addWidget(info_group)
        # Materials group
        materials_group = QGroupBox("Materials by Response Code")
        materials_layout = QGridLayout(materials_group)
        self.material_combos = {}
        current_materials = section.materials if hasattr(section, 'materials') else {}
        for row, code in enumerate(self.response_codes):
            code_label = QLabel(f"{code}:")
            combo = QComboBox()
            setup_material_dropdown(combo, lambda m: m.material_type == 'uniaxialMaterial', "Select Material")
            if code in current_materials:
                set_combo_to_material(combo, current_materials[code])
            combo.setToolTip(f"Select uniaxial material for response code {code}")
            materials_layout.addWidget(code_label, row, 0)
            materials_layout.addWidget(combo, row, 1)
            self.material_combos[code] = combo
        left_layout.addWidget(materials_group)
        # Base section group (optional)
        base_section_group = QGroupBox("Base Section (Optional)")
        base_section_layout = QFormLayout(base_section_group)
        self.base_section_combo = QComboBox()
        self.base_section_combo.addItem("None", None)
        for tag, s in Section.get_all_sections().items():
            self.base_section_combo.addItem(f"{s.user_name} (Tag: {tag})", s)
        if hasattr(section, 'base_section') and section.base_section:
            set_combo_to_material(self.base_section_combo, section.base_section)
        base_section_layout.addRow("Base Section:", self.base_section_combo)
        left_layout.addWidget(base_section_group)
        btn_layout = QHBoxLayout()
        save_btn = QPushButton("Save Changes")
        save_btn.clicked.connect(self.save_changes)
        cancel_btn = QPushButton("Cancel")
        cancel_btn.clicked.connect(self.reject)
        btn_layout.addWidget(save_btn)
        btn_layout.addWidget(cancel_btn)
        left_layout.addLayout(btn_layout)
        left_layout.addStretch()
        main_layout.addWidget(left_widget)
        line = QFrame()
        line.setFrameShape(QFrame.VLine)
        line.setFrameShadow(QFrame.Sunken)
        main_layout.addWidget(line)
        right_widget = QWidget()
        right_layout = QVBoxLayout(right_widget)
        help_display = QTextBrowser()
        help_display.setHtml(section.__class__.get_help_text())
        right_layout.addWidget(help_display)
        main_layout.addWidget(right_widget)
        main_layout.setStretch(0, 3)
        main_layout.setStretch(2, 2)
    def save_changes(self):
        try:
            materials = {}
            for code, combo in self.material_combos.items():
                material = get_material_by_combo_selection(combo)
                if material:
                    materials[code] = material
            if not materials:
                QMessageBox.warning(self, "Input Error", "Please select at least one material for a response code.")
                return
            base_section = self.base_section_combo.currentData()
            update_values = {"materials": materials}
            if base_section:
                update_values["base_section"] = base_section
            self.section.materials = materials
            if base_section:
                self.section.base_section = base_section
            QMessageBox.information(self, "Success", f"Section '{self.section.user_name}' updated successfully!")
            self.accept()
        except Exception as e:
            QMessageBox.critical(self, "Error", f"Failed to update section: {str(e)}")


if __name__ == "__main__":
    from femora.components.Material.materialsOpenSees import ElasticUniaxialMaterial, ElasticIsotropicMaterial
    from qtpy.QtWidgets import QApplication
    elas1 = ElasticUniaxialMaterial(user_name="Elastic1", E=30000, G=12000, nu=0.2)
    elas2 = ElasticUniaxialMaterial(user_name="Elastic2", E=25000, G=10000, nu=0.3)
    elas3 = ElasticIsotropicMaterial(user_name="Elastic3", E=20000, G=8000, nu=0.25)
    app = QApplication([])
    dialog = AggregatorSectionCreationDialog()
    if dialog.exec_() == QDialog.Accepted:
        print("Section created:", dialog.created_section)
    else:
        print("Creation cancelled")
    


    
    app.exec_()



from abc import ABC
from typing import List, Dict, Optional
from femora.components.event.event_bus import FemoraEvent, EventBus

class InterfaceBase(ABC):
    """Common logic for all interface objects."""

    _registry: dict[str, "InterfaceBase"] = {}

    def __init__(self, name: str, owners: List[str]):
        if name in InterfaceBase._registry:
            raise ValueError(f"Interface with name '{name}' already exists")
        self.name = name
        self.owners = owners  # could be meshpart names or other ids
        InterfaceBase._registry[name] = self
        self._subscribe_events()

    # ------------------------------------------------------------------
    # Event handling
    # ------------------------------------------------------------------
    def _subscribe_events(self):
        EventBus.subscribe(FemoraEvent.PRE_ASSEMBLE, self._on_pre_assemble)
        EventBus.subscribe(FemoraEvent.POST_ASSEMBLE, self._on_post_assemble)
        EventBus.subscribe(FemoraEvent.RESOLVE_CORE_CONFLICTS, self._on_resolve_core_conflicts)
        EventBus.subscribe(FemoraEvent.PRE_EXPORT, self._on_pre_export)
        EventBus.subscribe(FemoraEvent.POST_EXPORT, self._on_post_export)

    # The following are intentionally no-ops; subclasses/mix-ins may override
    def _on_pre_assemble(self, **payload):
        pass

    def _on_post_assemble(self, **payload):
        pass

    def _on_pre_export(self, **payload):
        pass

    def _on_post_export(self, **payload):
        pass

    def _on_resolve_core_conflicts(self, **payload):
        pass
    # ------------------------------------------------------------------
    # Helper class-methods
    # ------------------------------------------------------------------
    @classmethod
    def get(cls, name: str):
        return cls._registry.get(name)

    @classmethod
    def all(cls):
        return cls._registry.copy() 
    


class InterfaceManager:
    """Singleton that keeps track of all InterfaceBase objects."""

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance
    
    def __init__(self):
        from femora.components.interface.embedded_beam_solid_interface import EmbeddedBeamSolidInterface
        self.beam_solid_interface = EmbeddedBeamSolidInterface


    # Public helpers ----------------------------------------------------
    def create_interface(self, interface_cls, *args, **kwargs) -> InterfaceBase:
        """Helper that instantiates *interface_cls* and returns it."""
        if not issubclass(interface_cls, InterfaceBase):
            raise TypeError("interface_cls must be a subclass of InterfaceBase")
        return interface_cls(*args, **kwargs)

    def get(self, name: str) -> Optional[InterfaceBase]:
        return InterfaceBase.get(name)

    def all(self) -> Dict[str, InterfaceBase]:
        return InterfaceBase.all()


from qtpy.QtWidgets import (
    QWidget, QVBoxLayout, QFormLayout, QDoubleSpinBox, QComboBox, QLabel, QTabWidget
)
from femora.components.transformation.transformation import GeometricTransformation2D, GeometricTransformation3D

class TransformationWidget2D(QWidget):
    """
    Widget for creating a 2D geometric transformation object.
    """
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setLayout(QVBoxLayout())
        form = QFormLayout()
        self.type_combo = QComboBox()
        self.type_combo.addItems(["Linear", "PDelta", "Corotational"])
        form.addRow(QLabel("Type:"), self.type_combo)
        self.d_xi = QDoubleSpinBox(); self.d_xi.setRange(-1e6, 1e6)
        self.d_yi = QDoubleSpinBox(); self.d_yi.setRange(-1e6, 1e6)
        self.d_xj = QDoubleSpinBox(); self.d_xj.setRange(-1e6, 1e6)
        self.d_yj = QDoubleSpinBox(); self.d_yj.setRange(-1e6, 1e6)
        form.addRow("dXi:", self.d_xi)
        form.addRow("dYi:", self.d_yi)
        form.addRow("dXj:", self.d_xj)
        form.addRow("dYj:", self.d_yj)
        self.layout().addLayout(form)

    def get_transformation(self):
        return GeometricTransformation2D(
            self.type_combo.currentText(),
            self.d_xi.value(),
            self.d_yi.value(),
            self.d_xj.value(),
            self.d_yj.value()
        )

class TransformationWidget3D(QWidget):
    """
    Widget for creating a 3D geometric transformation object.
    """
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setLayout(QVBoxLayout())
        form = QFormLayout()
        self.type_combo = QComboBox()
        self.type_combo.addItems(["Linear", "PDelta", "Corotational"])
        form.addRow(QLabel("Type:"), self.type_combo)
        self.vecxz_x = QDoubleSpinBox(); self.vecxz_x.setRange(-1e6, 1e6); self.vecxz_x.setValue(0)
        self.vecxz_y = QDoubleSpinBox(); self.vecxz_y.setRange(-1e6, 1e6); self.vecxz_y.setValue(0)
        self.vecxz_z = QDoubleSpinBox(); self.vecxz_z.setRange(-1e6, 1e6); self.vecxz_z.setValue(-1)
        self.d_xi = QDoubleSpinBox(); self.d_xi.setRange(-1e6, 1e6)
        self.d_yi = QDoubleSpinBox(); self.d_yi.setRange(-1e6, 1e6)
        self.d_zi = QDoubleSpinBox(); self.d_zi.setRange(-1e6, 1e6)
        self.d_xj = QDoubleSpinBox(); self.d_xj.setRange(-1e6, 1e6)
        self.d_yj = QDoubleSpinBox(); self.d_yj.setRange(-1e6, 1e6)
        self.d_zj = QDoubleSpinBox(); self.d_zj.setRange(-1e6, 1e6)
        form.addRow("vecXZ X:", self.vecxz_x)
        form.addRow("vecXZ Y:", self.vecxz_y)
        form.addRow("vecXZ Z:", self.vecxz_z)
        form.addRow("dXi:", self.d_xi)
        form.addRow("dYi:", self.d_yi)
        form.addRow("dZi:", self.d_zi)
        form.addRow("dXj:", self.d_xj)
        form.addRow("dYj:", self.d_yj)
        form.addRow("dZj:", self.d_zj)
        self.layout().addLayout(form)

    def get_transformation(self):
        return GeometricTransformation3D(
            self.type_combo.currentText(),
            self.vecxz_x.value(),
            self.vecxz_y.value(),
            self.vecxz_z.value(),
            self.d_xi.value(),
            self.d_yi.value(),
            self.d_zi.value(),
            self.d_xj.value(),
            self.d_yj.value(),
            self.d_zj.value()
        )
    
    def load_transformation(self, transformation):
        if transformation is None:
            return
        self.type_combo.setCurrentText(getattr(transformation, 'transf_type', 'Linear'))
        self.vecxz_x.setValue(getattr(transformation, 'vecxz_x', 0))
        self.vecxz_y.setValue(getattr(transformation, 'vecxz_y', 0))
        self.vecxz_z.setValue(getattr(transformation, 'vecxz_z', -1))
        self.d_xi.setValue(getattr(transformation, 'd_xi', 0))
        self.d_yi.setValue(getattr(transformation, 'd_yi', 0))
        self.d_zi.setValue(getattr(transformation, 'd_zi', 0))
        self.d_xj.setValue(getattr(transformation, 'd_xj', 0))
        self.d_yj.setValue(getattr(transformation, 'd_yj', 0))
        self.d_zj.setValue(getattr(transformation, 'd_zj', 0))

# Optionally, you can keep the following demo for manual testing:
if __name__ == "__main__":
    from qtpy.QtWidgets import QApplication
    import sys
    app = QApplication(sys.argv)
    tabs = QTabWidget()
    tabs.addTab(TransformationWidget2D(), "2D Transformation")
    tabs.addTab(TransformationWidget3D(), "3D Transformation")
    tabs.show()
    sys.exit(app.exec_())